#ifndef FUNCTION_TYPE_MD
#define FUNCTION_TYPE_MD

/**
 * @file FunctionType1D.h
 * @author Bryan BERTHOU (SPhN / CEA Saclay)
 * @date October 07, 2016
 * @version 1.0
 */

#include <vector>

#include "../../linear_algebra/vector/VectorD.h"

namespace NumA {

/**
 * @class FunctionTypeMD
 *
 * @brief Class for defining multi-dimensional functions that can be used as arguments in virtual methods.
 * This is needed to circumvent the issue that templates and virtual methods are incompatible.
 *
 * Usage:
 *
 * ```cpp
 * NumA::FunctionTypeMD* myFunctor = NumA::FunctorUtils::newFunctorMD(this,
 *                                   &MyClass::myFunction);
 * ```
 * where `myFunction` is some function defined in some class `MyClass` with the following signature:
 * ```cpp
 * double MyClass::myFunction(NumA::VectorD& variables, std::vector<double>& parameters);
 * ```
 *
 * The pointer `this` needs to be replaced by a pointer to the object instantiating `MyClass`, if the first line is not written inside `MyClass`.
 *
 * Now, you can pass `myFunctor` as argument to any method (e.g. integration routines), even virtual ones:
 * ```cpp
 * virtual SomeReturnType someVirtualMethod(..., NumA::FunctionTypeMD* functor, ...);
 * ```
 * and use the functor as a function (with the `operator()`) inside this method:
 * ```cpp
 * functor(variables, parameters); // parameters is optional.
 * ```
 *
 * See also FunctorUtils documentation.
 *
 * For one dimensional functions, see FunctionType1D.
 */

class FunctionTypeMD {
public:
    /**
     * Empty constructor.
     */
    FunctionTypeMD() {
    }

    /**
     * Default destructor.
     */
    virtual ~FunctionTypeMD() {
    }

    /**
     * Operator allowing the functor to be used as a function.
     * @param variables Multi-variate argument of the function represented by the functor.
     * @param parameters Parameters that can be passed to the function.
     * @return double
     */
    virtual double operator()(VectorD &variables,
            std::vector<double> &parameters) = 0;
    /**
     * Operator allowing the functor to be used as a function.
     * @param variables Multi-variate argument of the function represented by the functor.
     * @return double
     */
    virtual double operator()(VectorD &variables) = 0;
};

} // namespace NumA

#endif /* FUNCTION_TYPE_MD */
