#ifndef FUNCTION_TYPE_1D
#define FUNCTION_TYPE_1D

/**
 * @file FunctionType1D.h
 * @author Bryan BERTHOU (SPhN / CEA Saclay)
 * @date 15 February 2016
 * @version 1.0
 */

#include <vector>

namespace NumA {

/**
 * @class FunctionType1D
 *
 * @brief Class for defining one-dimensional functions that can be used as arguments in virtual methods.
 * This is needed to circumvent the issue that templates and virtual methods are incompatible.
 *
 * Usage:
 *
 * ```cpp
 * NumA::FunctionType1D* myFunctor = NumA::FunctorUtils::newFunctor1D(this,
 *                                   &MyClass::myFunction);
 * ```
 * where `myFunction` is some function defined in some class `MyClass` with the following signature:
 * ```cpp
 * double MyClass::myFunction(double variable, std::vector<double>& parameters);
 * ```
 *
 * The pointer `this` needs to be replaced by a pointer to the object instantiating `MyClass`, if the first line is not written inside `MyClass`.
 *
 * Now, you can pass `myFunctor` as argument to any method (e.g. integration routines), even virtual ones:
 * ```cpp
 * virtual SomeReturnType someVirtualMethod(..., NumA::FunctionType1D* functor, ...);
 * ```
 * and use the functor as a function (with the `operator()`) inside this method:
 * ```cpp
 * functor(variable, parameters); // parameters is optional.
 * ```
 *
 * See also FunctorUtils documentation.
 *
 * See also Integrator1D documentation for the specific case of integration routines.
 *
 * For multi-dimensional functions, see FunctionTypeMD.
 */

class FunctionType1D {
public:
    /**
     * Empty constructor.
     */
    FunctionType1D() {
    }

    /**
     * Default destructor.
     */
    virtual ~FunctionType1D() {
    }

    /**
     * Operator allowing the functor to be used as a function.
     * @param x Argument of the function represented by the functor.
     * @param parameters Parameters that can be passed to the function.
     * @return double
     */
    virtual double operator()(double x, std::vector<double> &parameters) = 0;
    /**
     * Operator allowing the functor to be used as a function.
     * @param x Argument of the function represented by the functor.
     * @return double
     */
    virtual double operator()(double x) = 0;
};

} // namespace NumA

#endif /* FUNCTION_TYPE_1D */
