#ifndef INTEGRATOR_REGISTRY_H
#define INTEGRATOR_REGISTRY_H

/**
 * @file IntegratorRegistry.h
 * @author Bryan BERTHOU (SPhN / CEA Saclay)
 * @date June 29, 2016
 * @version 1.0
 */

#include <SFML/System/Mutex.hpp>
#include <map>
#include <string>

#include "Integrator.h"

namespace NumA {

/**
 * @class IntegratorRegistry
 *
 * @brief
 */
class IntegratorRegistry {
public:
    /**
     * Static function to be able to retrieve a unique instance pointer of this class anywhere in the code.
     *
     * @return a unique instance of this class
     */
    static IntegratorRegistry* getInstance();

    /**
     * Default destructor
     */
    virtual ~IntegratorRegistry();

    unsigned int registerIntegrator(Integrator * pIntegrator);

    /**
     * Try to find stored object identified by its unique class identifier.
     *
     * @param classId
     * @return reference to object in memory or throw an exception
     */
    Integrator* get(unsigned int classId) const;

    /**
     * Try to find stored object identified by its class name.
     *
     * @param className
     * @return reference to object in memory or throw an exception
     */
    Integrator* get(const std::string &className) const;

private:

    /**
     * Because NumA++ can be use in a multi-threading program we must ensure that only one object manipulates the registry at the same time to avoid memory access violation.
     * It's performed by a mechanism called mutex or semaphore.
     */
    mutable sf::Mutex m_mutex;

    /**
     * Private pointer of this class for a unique instance
     */
    static IntegratorRegistry* m_pInstance;

    /**
     * Private default constructor for a unique instance
     */
    IntegratorRegistry();

    std::map<unsigned int, Integrator*> m_baseObjectList; ///< list of registered objects identified by their unique integer identifier
    std::map<std::string, Integrator*> m_translateList; ///< list of registered objects identified by their class name.
};

} // namespace NumA

#endif /* INTEGRATOR_REGISTRY_H */
