#ifndef INTEGRATOR_1D_H
#define INTEGRATOR_1D_H

/**
 * @file Integrator1D.h
 * @author Bryan BERTHOU (SPhN / CEA Saclay)
 * @date February 15, 2016
 * @version 1.0
 */

#include <string>
#include <vector>

#include "../../functor/one_dimension/Functor1D.h"
#include "../../utils/Errors.h"
#include "../../utils/Tolerances.h"
#include "IntegratorType1D.h"

namespace ElemUtils {
class Parameters;
} /* namespace ElemUtils */

namespace NumA {

/**
 * @class Integrator1D
 *
 * @brief Abstract class for all integration routines.
 *
 * Example:
 * ```cpp
 * NumA::Integrator1D* integrator = NumA::Integrator1D::newIntegrator(NumA::IntegratorType1D::GaussLegendre); // Using Gauss-Legendre quadrature.
 * NumA::FunctionType1D* functor = NumA::FunctorUtils::newFunctor1D(pointerMyClass, &MyClass::myFunction); // See FunctionType1D documentation.
 * double integral = integrator.integrate(functor, -1., 1.); // We integrate a function on [-1,1].
 * ```
 */

class Integrator1D {
public:
    static const std::string PARAM_NAME_ABSOLUTE_TOLERANCE; ///< Parameter used in the configure() to set the absolute tolerance.
    static const std::string PARAM_NAME_RELATIVE_TOLERANCE; ///< Parameter used in the configure() to set the relative tolerance.

    /**
     * Default constructor.
     */
    Integrator1D();

    /**
     * Default destructor.
     */
    virtual ~Integrator1D();

    virtual Integrator1D* clone() const = 0;

    /**
     * Integration routine.
     * @param pFunction Functor representing the one-dimensional function to integrate.
     * @param a Lower bound.
     * @param b Upper bound.
     * @param parameters Parameters that can be passed to the function.
     * @return Integral.
     */
    virtual double integrate(FunctionType1D* pFunction, double a, double b,
            std::vector<double> &parameters) = 0;

    /**
     * Provides a generic method to configure all types of integrations by passing a Parameters object.
     * Parameters class represents a list of couples key/value (see Parameters class documentation for more info).
     *
     * @param parameters ElemUtils::Parameters object.
     */
    virtual void configure(const ElemUtils::Parameters &parameters);

    //TODO moved to FunctorUtils ; remove this definition
    /**
     * Use FunctorUtils::newFunctor1D instead.
     *
     * TODO: Remove this method.
     */
    template<typename PointerToObj, typename PointerToMemFn>
    static Functor1D<PointerToObj, PointerToMemFn>* newIntegrationFunctor(
            PointerToObj* object, PointerToMemFn function) {
        return new Functor1D<PointerToObj, PointerToMemFn>(object, function);
    }

    /**
     * Instantiates an Integrator object.
     * Will be replaced by a Factory/Registry system in the future.
     * @param oneDimIntegratorType Type of integrator.
     * @return Integrator1D pointer.
     */
    static Integrator1D* newIntegrator(
            const IntegratorType1D::Type &oneDimIntegratorType);

    // ##### GETTERS & SETTERS #####

    /**
     *
     * @return Absolute and relative tolerances.
     */
    const Tolerances& getTolerances() const;
    /**
     *
     * @param tolerances Absolute and relative tolerances.
     */
    void setTolerances(const Tolerances& tolerances);
    /**
     *
     * @return Absolute and relative errors estimations.
     */
    const Errors& getErrors() const;
    /**
     *
     * @param errors Absolute and relative errors estimations.
     */
    void setErrors(const Errors& errors);

protected:
    Tolerances m_tolerances; ///< Absolute and relative tolerances.
    Errors m_errors; ///< Absolute and relative errors estimations.

    /**
     * Copy constructor.
     * Called by clone().
     * @param other Integrator1D object to copy.
     */
    Integrator1D(const Integrator1D &other);
};

} // namespace NumA

#endif /* INTEGRATOR_1D_H */
