#ifndef INTEGRATOR_TYPE_1D_H
#define INTEGRATOR_TYPE_1D_H

/**
 * @file IntegratorType1D.h
 * @author Bryan BERTHOU (SPhN / CEA Saclay)
 * @date 17 February 2016
 * @version 1.0
 */

#include <string>

namespace NumA {

/**
 * @class IntegratorType1D
 *
 * @brief Type of one-dimensional integrations (wrapper for enum).
 *
 * See Integrator1D documentation for an example.
 */

class IntegratorType1D {

public:

	/**
	 * Type of integration.
	 */
	enum Type {
		UNDEFINED = 0, //!< Undefined type.
		TRAPEZOIDAL = 1, //!< Trapezoidal method with a fixed number of equidistant steps.
		TRAPEZOIDALLOG = 2, //!< Trapezoidal method with a fixed number with equidistantly distributed steps in [ln(min), ln(max)] range.
		GL = 3, //!< Gauss-Legendre method with a fixed number of steps.
		DEXP = 4, //!< Adaptive double exponential method.
		GK21_ADAPTIVE = 5  //!< Adaptive Gauss-Kronrod method with 21 point rules.
	};

	/**
	 * Default constructor.
	 */
	IntegratorType1D();

	/**
	 * Assignment constructor.
	 * @param type Type to be assigned.
	 */
	IntegratorType1D(Type type);

	/**
	 * Assignment constructor trying to match integration type from given string. If unable to match set IntegratorType1D::UNDEFINED.
	 * @param integratorType1DString String to be matched.
	 */
	IntegratorType1D(const std::string& integratorType1DString);

	/**
	 * Default destructor.
	 */
	virtual ~IntegratorType1D();

	/**
	 * Operator that casts back into an enum.
	 */
	operator Type() const;

	/**
	 * Returns a formatted string displaying the type.
	 * @return std::string
	 */
	std::string toString() const;

	/**
	 * Returns a short name for the type.
	 * @return std::string
	 */
	std::string getShortName();

	/**
	 * Returns the wrapped enum.
	 * @return IntegratorType1D::Type
	 */
	IntegratorType1D::Type getType() const;

	/**
	 * Set the type of integration.
	 * @param type Integration type to be set.
	 */
	void setType(Type type);

private:
	IntegratorType1D::Type m_type; ///< Type of integration (enum).
};

} /* namespace NumA */

#endif /* INTEGRATOR_TYPE_1D_H */
