/**
 * @file Chebyshev.h
 * @author Nabil Chouika (Irfu/SPhN - CEA Saclay)
 * @date 22 mars 2016
 * @version 1.0
 */

#ifndef CHEBYSHEV_H_
#define CHEBYSHEV_H_

#include "../linear_algebra/matrix/MatrixD.h"
#include "../linear_algebra/vector/VectorD.h"

namespace NumA {

/**
 * @class Chebyshev
 * @brief Chebyshev expansion class.
 *
 * Interpolation \f$ f(x) = \sum_{i=0}^{N-1} c_i T_i \left( x \right) \f$ for \f$ x \in \left[ -1, 1 \right] \f$ ,
 * where \f$ N \f$ is the order of the expansion.
 *
 * TODO: \f$ x \in \f$ any interval.
 */

class Chebyshev {
public:
    /**
     * Default Constructor with the order N of the expansion.
     * @param N Order of the expansion.
     */
    Chebyshev(unsigned int N = 50);
    /**
     * Default destructor.
     */
    virtual ~Chebyshev();

    /**
     * Get the order N of the expansion.
     * @return N unsigned int
     */
    unsigned int getN() const;
    /**
     * Set the order N of the expansion/
     * @param n unsigned int
     */
    void setN(unsigned int n);

    const NumA::VectorD& getRoots() const;

    /**
     * Interpolation matrix.
     * @return Matrix which gives a vector of the coefficients for the expansion when applied to a vector of values on Chebyshev nodes.
     */
    const NumA::MatrixD& getValuesToCoeffsMatrix() const;
    /**
     *
     * @return Matrix which gives a vector of the values on Chebyshev nodes when applied to a vector of coefficients for the expansion.
     */
    const NumA::MatrixD& getCoeffsToValuesMatrix() const;

    /**
     *
     * @param points Points where to evaluate the expansion.
     * @return Matrix which gives a vector of the values on given points when applied to a vector of coefficients for the expansion.
     */
    NumA::MatrixD getCoeffsToValuesMatrix(const NumA::VectorD& points) const;
    /**
     *
     * @param x Point where the expansion is evaluated.
     * @return Vector which gives the Chebyshev expansion on x when a scalar product is taken with a vector of coefficients for the expansion.
     */
    NumA::VectorD getCoeffsToValueVector(double x) const;

    /**
     * Interpolation matrix.
     * @param points Points where to evaluate the expansion.
     * @return Matrix which gives a vector of values on given points when applied to a vector of values on Chebyshev nodes.
     */
    NumA::MatrixD getValuesToValuesMatrix(const NumA::VectorD& points) const;
    /**
     *
     * @param x Point where the expansion is evaluated.
     * @return Vector which gives the Chebyshev expansion on x when a scalar product is taken with a vector of values on Chebyshev nodes.
     */
    NumA::VectorD getValuesToValueVector(double x) const;

    /**
     * Evaluate the Chebyshev expansion.
     * @param coeffs Coefficients \f$ c_i \f$ of the expansion.
     * @param x Point where the expansion is evaluated.
     * @return Chebyshev expansion: \sum_{i=0}^{N-1} c_i T_i \left( x \right) \f$.
     */
    double evaluateExpansion(const NumA::VectorD& coeffs, double x) const;

    static double T(unsigned int i, double x); ///< \f$ T_i \left( x \right) = \cos \left( i \arccos \left( x \right) \right) \f$.

private:
    unsigned int m_N; ///< Order of the expansion.
    NumA::VectorD m_roots; ///< Roots of T_N, i.e. the nodes of the expansion.
    NumA::MatrixD m_valuesToCoeffsMatrix; ///< Transformation matrix from the values on nodes to the coefficients of the expansion.
    NumA::MatrixD m_coeffsToValuesMatrix; ///< Transformation matrix from the coefficients of the expansion to the values on nodes.

    void initRoots(); ///< Initialize the roots.
    void initValuesToCoeffsMatrix(); ///< Initialize the interpolation matrix.
};

} /* namespace NumA */

#endif /* CHEBYSHEV_H_ */
