/**
 * @file LSMRSolver.h
 * @author Nabil Chouika (Irfu/SPhN - CEA Saclay)
 * @date 2 déc. 2016
 * @version 1.0
 */

#ifndef LSMRSOLVER_H_
#define LSMRSOLVER_H_

#include <string>
#include <vector>
#include <cstddef>

#include "../matrix/MatrixD.h"
#include "../vector/VectorD.h"

namespace ElemUtils {
class Formatter;
} /* namespace ElemUtils */

namespace NumA {

/**
 * @class LSMRSolver
 * @brief LSMR solves Ax = b or min ||Ax - b|| with or without damping,
 * using the iterative algorithm of David Fong and Michael Saunders:
 *     http://www.stanford.edu/group/SOL/software/lsmr.html
 *
 * Adapted for C++ from the scipy implementation:
 * https://github.com/scipy/scipy/blob/v0.18.1/scipy/sparse/linalg/isolve/lsmr.py
 */

class LSMRSolver {
public:
    /**
     * Constructor.
     * @param damp Damping factor for regularized least-squares.
     * @param atol Stopping tolerance.
     * @param btol Stopping tolerance.
     * @param conlim LSMR terminates if an estimate of `cond(A)` exceeds conlim.
     * @param maxiter LSMR terminates if the number of iterations reaches maxiter.
     * @param output Pointer to the stream where to write the output.
     */
    LSMRSolver(double damp = 0., double atol = 1.e-6, double btol = 1.e-6,
            double conlim = 1.e8, size_t maxiter = 0,
            ElemUtils::Formatter* output = 0);
    /**
     * Default destructor.
     */
    virtual ~LSMRSolver();

    /**
     * @brief Solves the linear system \f$ A X = B \f$ and returns \f$ X \f$ in the least-squares sense.
     * Iterative method for large sparse matrices.
     * @param A MatrixD.
     * @param B VectorD.
     * @return VectorD Solution X.
     */
    VectorD solve(const MatrixD & A, const VectorD & B);

    // ##### GETTERS & SETTERS #####

    double getToleranceA() const; ///< Stopping tolerance. LSMR continues iterations until a certain backward error estimate is smaller than some quantity depending on atol and btol.
    void setToleranceA(double atol); ///< Stopping tolerance. LSMR continues iterations until a certain backward error estimate is smaller than some quantity depending on atol and btol.
    double getToleranceB() const; ///< Stopping tolerance. LSMR continues iterations until a certain backward error estimate is smaller than some quantity depending on atol and btol.
    void setToleranceB(double btol); ///< Stopping tolerance. LSMR continues iterations until a certain backward error estimate is smaller than some quantity depending on atol and btol.
    double getConditionNumberLimit() const; ///< LSMR terminates if an estimate of `cond(A)` exceeds conlim.
    void setConditionNumberLimit(double conlim); ///< LSMR terminates if an estimate of `cond(A)` exceeds conlim.
    double getDamping() const; ///< Damping factor for regularized least-squares.
    void setDamping(double damp); ///< Damping factor for regularized least-squares.
    size_t getMaximumIerations() const; ///< LSMR terminates if the number of iterations reaches maxiter.
    void setMaximumIerations(size_t maxiter); ///< LSMR terminates if the number of iterations reaches maxiter.
     /**
      *
      * @return Reference to the stream where the output is written.
      */
    ElemUtils::Formatter& getOutput() const;
    /**
     *
     * @param output Reference to the stream where the output is written.
     */
    void setOutput(ElemUtils::Formatter& output);
    double getConditionNumber() const;
    unsigned int getStoppingCase() const; ///< Reason for stopping the iterations.
    size_t getNumberIterations() const;  ///< Number of iterations.
    double getNormA() const;
    double getNormAr() const;
    double getNormr() const;
    double getNormX() const;
    const std::string& getStoppingMessage() const; ///< Reason for stopping the iterations.

private:
    // ##### Parameters of the solver #####
    double m_damp; ///< Damping factor for regularized least-squares.
    double m_atol; ///< Stopping tolerance. LSMR continues iterations until a certain backward error estimate is smaller than some quantity depending on atol and btol.
    double m_btol; ///< Stopping tolerance. LSMR continues iterations until a certain backward error estimate is smaller than some quantity depending on atol and btol.
    double m_conlim; ///< LSMR terminates if an estimate of `cond(A)` exceeds conlim.
    size_t m_maxiter; ///< LSMR terminates if the number of iterations reaches maxiter.
    ElemUtils::Formatter* m_output; ///< Pointer to the stream where to write the output.

    // ##### Results of the solver #####
    unsigned int m_istop; ///< Reason for stopping the iterations.
    size_t m_itn; ///< Number of iterations.
    double m_normr, m_normar, m_normA, m_condA, m_normx;
    std::string m_stopMessage; ///< Reason for stopping the iterations.

    // Miscellaneous
    std::vector<std::string> m_msg;
    std::string m_hdg1;
    std::string m_hdg2;
    unsigned int m_pfreq;   // print frequency (for repeating the heading)

    /**
     * Routine of rotation from scipy.
     * @param a
     * @param b
     * @param c
     * @param s
     * @param r
     */
    void _sym_ortho(double a, double b, double& c, double& s, double& r) const;
};

} /* namespace NumA */

#endif /* LSMRSOLVER_H_ */
