#ifndef MATRIX_D_H
#define MATRIX_D_H

/**
 * @file MatrixD.h
 * @author Bryan BERTHOU (Irfu/SPhN - CEA Saclay)
 * @author Nabil CHOUIKA (Irfu/SPhN - CEA Saclay)
 * @date 23 October 2014
 */

#include <string>
#include <vector>
#include <stddef.h>

#include "../vector/VectorD.h"

namespace NumA {

class VectorD;

/**
 * @class MatrixD
 *
 * @brief Represents a two-dimensional array of double.
 *
 * The indices start at 0 up to `size`-1.
 *
 * Examples:
 * ```cpp
 * NumA::MatrixD A(5,5); // 5x5 matrix with coefficients equal to 0.
 * A.at(1,3) = 10.; // Defines one coefficient (second row, fourth column).
 * NumA::MatrixD B = NumA::MatrixD::Id(5); // Identity matrix.
 * B.at(4,4) = 5.; // Last row, last column.
 * // Matrix operations:
 * NumA::MatrixD C;
 * C = A*B + B/2.;
 * C.toString();
 * ```
 * This returns:
 * ```
 * 0.5 0 0 0 0
 * 0 0.5 0 10 0
 * 0 0 0.5 0 0
 * 0 0 0 0.5 0
 * 0 0 0 0 2.5
 * ```
 */

class MatrixD {
public:
    /**
     * Default constructor. \n
     * Initializes an empty matrix.
     */
    MatrixD();

    /**
     * Creates a matrix with specific dimensions and fills it with given values separated by comma.
     *
     * TODO: Write something better?
     *
     * If the number and type of arguments are not right, UNDEFINED BEHAVIOUR !!!
     * This constructor is very DANGEROUS!!! USE IT RIGHT, OR SCRAP IT ALL AND REWRITE SOMETHING BETTER!
     *
     * @param _rowsNumber : number of rows
     * @param _columnsNumber : number of columns
     * @param first_value : first value of the matrix
     */
    MatrixD(const size_t _rowsNumber, const size_t _columnsNumber,
            double first_value, ...);

    /**
     * Creates matrix with specific dimensions and fills it with 0.
     *
     * @param _rowsNumber : number of rows
     * @param _columnsNumber : number of columns
     */
    MatrixD(const size_t _rowsNumber, const size_t _columnsNumber);

    /**
     * Copy constructor.
     *
     * @param rhs Matrix to be copied.
     */
    MatrixD(const MatrixD& rhs);

    /**
     * VectorD constructor (creates a one-column matrix).
     *
     * @param rhs Vector to be converted to a matrix.
     */
    MatrixD(const VectorD& rhs);

    /**
     * Default destructor.
     */
    virtual ~MatrixD();

    /**
     * Assigns a new matrix with all coefficients set to `value`.
     * @param _rowsNumber Number of rows.
     * @param _columnsNumber Number of columns.
     * @param value Default value for all the coefficients.
     */
    void assign(const size_t _rowsNumber, const size_t _columnsNumber,
            double value = 0.);

    /**
     * Sets an existing line of the matrix to new values given by a vector.
     * @param i Index of the line to be modified.
     * @param line Vector to copy in the matrix.
     */
    void setLine(size_t i, const NumA::VectorD& line);
    /**
     * Adds a new line in the matrix (its dimension is therefore incremented) with the values given by a vector.
     * @param i Index the row where the line will be inserted.
     * @param line Vector to copy in the matrix.
     */
    void addLine(size_t i, const NumA::VectorD& line);
    /**
     * Adds a new line at the end of the matrix (its dimension is therefore incremented) with the values given by a vector.
     * @param line Vector to copy in the matrix.
     */
    void appendLine(const NumA::VectorD& line);
    /**
     * Extracts a single line from the matrix.
     * @param lineIndex Index of the row to be extracted.
     * @return Vector corresponding to the row.
     */
    NumA::VectorD getLine(const size_t lineIndex) const;

    // Matrix/vector operations
    /**
     * Matrix vector multiplication.
     * @param rhs Vector.
     * @return Vector \f$ M \cdot V \f$.
     */
    NumA::VectorD operator*(const NumA::VectorD& rhs) const;
    //std::vector<double> operator*=(const std::vector<double>& rhs);

    // Matrix/matrix operations
    /**
     * Matrix-matrix multiplication.
     * @param rhs Matrix.
     * @return Product of matrices.
     */
    NumA::MatrixD operator*(const NumA::MatrixD& rhs) const;
    /**
     * Matrix addition.
     * @param rhs Matrix.
     * @return Sum of matrices.
     */
    NumA::MatrixD operator+(const NumA::MatrixD& rhs) const;
    /**
     * Matrix subtraction.
     * @param rhs Matrix.
     * @return Difference of matrices.
     */
    NumA::MatrixD operator-(const NumA::MatrixD& rhs) const;

    // Matrix/double operations
    /**
     * Multiplication of all the coefficients by a scalar.
     * @param rhs Scalar.
     * @return Matrix.
     */
    NumA::MatrixD operator*(double rhs) const;
    /**
     * Addition coefficient by coefficient with a scalar.
     * @param rhs Scalar.
     * @return Matrix.
     */
    NumA::MatrixD operator+(double rhs) const;
    /**
     * Subtraction coefficient by coefficient with a scalar.
     * @param rhs Scalar.
     * @return Matrix.
     */
    NumA::MatrixD operator-(double rhs) const;
    /**
     * Division of all the coefficients by a scalar.
     * @param rhs Scalar.
     * @return Matrix.
     */
    NumA::MatrixD operator/(double rhs) const;

    /**
     * Matrix transpose.
     * @return Matrix transposed.
     */
    NumA::MatrixD transpose() const;

    /**
     * Matrix identity.
     * @param n Dimension.
     * @return Matrix identity.
     */
    static NumA::MatrixD Id(unsigned int n);

    /**
     * Update the value at the given indices.
     *
     * @param i Row index.
     * @param j Column index.
     * @param value New value for the coefficient.
     */
    void update(const size_t i, const size_t j, const double value);

    /**
     * Element access.
     * Can be used to set the value of a coefficient.
     * @param i Row index.
     * @param j Column index.
     * @return Reference to a certain coefficient of the matrix.
     */
    double & at(const size_t i, const size_t j);
    /**
     * Element access.
     * @param i Row index.
     * @param j Column index.
     * @return Constant reference to a certain coefficient of the matrix.
     */
    const double & at(const size_t i, const size_t j) const;

    /**
     * Return a formatted characters string to display matrix's values.
     *
     * @return std::string
     */
    std::string toString() const;

    // ##### GETTERS & SETTERS #####
    /**
     *
     * @return Number of columns.
     */
    size_t cols() const;
    /**
     *
     * @return Number of rows.
     */
    size_t rows() const;

private:
    std::vector<double> m_matrix; ///< A flat std::vector to represent the matrix.
    size_t m_rowsNumber;    ///< Number of rows.
    size_t m_columnsNumber;    ///< Number of columns.
};

} /* namespace NumA */

#endif /* MATRIX_D_H */
