/*
 * MatrixComplexD.h
 *
 *  Created on: Jan 16, 2016
 *      Author: Pawel Sznajder (IPNO)
 */

#ifndef MATRIXCOMPLEXD_H_
#define MATRIXCOMPLEXD_H_

#include <complex>
#include <vector>

namespace NumA {

/**
 * @class MatrixComplexD
 *
 * @brief Matrix of complex numbers of undefined size.
 *
 * This class represents a matrix of complex double precision numbers. The size of this matrix should be defined by the user.
 */
class MatrixComplexD {

public:

	/**
	 * Default constructor.
	 */
	MatrixComplexD();

	/**
	 * Assignemnt constructor.
	 * Elements will be initialized with `std::complex<double>(0., 0.).`
	 * @param n Size of x dim. to be set.
	 * @param m Size of y dim. to be set.
	 */
	MatrixComplexD(const unsigned int n, const unsigned int m);

	/**
	 * Copy constructor
	 * @param v Object to be copied.
	 */
	MatrixComplexD(const MatrixComplexD& v);

	/**
	 * Destructor.
	 */
	virtual ~MatrixComplexD();

	/**
	 * Dimension index.
	 */
	enum Dim {
		X = 1, 	///< Dim. x.
		Y = 2	///< Dim. y.
	};

	/**
	 * Get conjunction of matrix.
	 * Return: \f$M^{*}\f$.
	 */
	MatrixComplexD Conjunct() const;

	/**
	 * Multiply this with other matrix.
	 * Return: \f$M N\f$.
	 * @param v Other matrix,
	 */
	MatrixComplexD Mult(const MatrixComplexD& v) const;

	/**
	 * Clear vector (set `std::complex<double>(0., 0.)` to all elements).
	 */
	void Clear();

	/**
	 * Make unit (set `std::complex<double>(1., 0.)` to diagonal elements).
	 */
	void MakeUnit();

	/**
	 * Operator =.
	 */
	void operator=(const MatrixComplexD& rhs);

	/**
	 * Operator +.
	 */
	MatrixComplexD operator+(const MatrixComplexD& rhs);

	/**
	 * Operator -.
	 */
	MatrixComplexD operator-(const MatrixComplexD& rhs);

	/**
	 * Operator * by real number (rhs).
	 */
	MatrixComplexD operator*(const double& rhs);

	/**
	 * Operator * by real number (lhs).
	 */
	friend MatrixComplexD operator*(const double& lhs, MatrixComplexD& rhs);

	/**
	 * Operator * by complex number (rhs).
	 */
	MatrixComplexD operator*(const std::complex<double>& rhs);

	/**
	 * Operator * by complex number (lhs).
	 */
	friend MatrixComplexD operator*(const std::complex<double>& lhs,
			MatrixComplexD& rhs);

	/**
	 * Get dim. size for given dimension.
	 * @param v Dimension.
	 */
	unsigned int GetNElements(const Dim v) const;

	/**
	 * Set element
	 * @param i Index of x element to be set.
	 * @param j Index of y element to be set.
	 * @param value Value to be set.
	 */
	void SetElement(const unsigned int i, const unsigned int j,
			const std::complex<double> value);

	/**
	 * Get element.
	 * @param i Index of x element to be retrieved.
	 * @param j Index of y element to be retrieved.
	 */
	std::complex<double> GetElement(const unsigned int i,
			const unsigned int j) const;

private:

	/**
	 * Compare number of elements of this and other vector.
	 * @param v Other matrix.
	 */
	void CompareNElements(const MatrixComplexD& v) const;

	/**
	 * Size of x dim.
	 */
	unsigned int m_n;

	/**
	 * Size of y dim.
	 */
	unsigned int m_m;

	/**
	 * Elements
	 */
	std::vector<std::vector<std::complex<double> > > m_elements;
};

} /* namespace NumA */

#endif /* MATRIXCOMPLEXD_H_ */
