#ifndef VECTOR_D_H
#define VECTOR_D_H

/**
 * @file VectorD.h
 * @author Bryan BERTHOU (SPhN / CEA Saclay)
 * @date 07 September 2015
 * @version 1.0
 */

#include <string>
#include <vector>
#include <stddef.h>

//#include "../matrix/MatrixD.h"

namespace NumA {
class Vector2D;
class Vector3D;
class Vector4D;
} /* namespace NumA */

namespace NumA {

class MatrixD;

/**
 * @class VectorD
 *
 * @brief Object representing a mathematical vector.
 */

class VectorD {
public:
    /**
     * Default constructor.
     */
    VectorD();

    /**
     * Construct a vector of given size initialized at 0.
     * @param size Size of the vector.
     */
    VectorD(size_t size);

    /**
     * Construct a vector of given size initialized at initValue.
     * @param size Size of the vector.
     * @param initValue Default value for the coefficients.
     */
    VectorD(size_t size, double initValue);

    /**
     * Copy constructor.
     * @param vector std::vector<double>
     */
    VectorD(const std::vector<double> &vector);

    /**
     * Copy constructor.
     * @param vector VectorD
     */
    VectorD(const VectorD &vector);

    /**
     * Copy constructor.
     * @param vector Vector2D
     */
    VectorD(const Vector2D &vector);

    /**
     * Copy constructor.
     * @param vector Vector3D
     */
    VectorD(const Vector3D &vector);

    /**
     * Copy constructor.
     * @param vector Vector4D
     */
    VectorD(const Vector4D &vector);

    /**
     * Default destructor.
     */
    virtual ~VectorD();

    void push_back(double value); ///< Same as std::vector.
    void resize(size_t n, double value = 0.); ///< Same as std::vector.
    void assign(size_t n, double value = 0.); ///< Same as std::vector.

    /**
     * @return size_t Size of the vector.
     */
    size_t size() const;

    /**
     * Returns a subset of the vector: [startPos:endPos[
     * @param startPos
     * @param endPos
     * @return VectorD
     */
    VectorD sub(size_t startPos, size_t endPos) const;

    /**
     * Scalar product.
     * @param rhs VectorD of same size.
     * @return double
     */
    double operator *(const VectorD &rhs) const;

    /**
     * Subtraction.
     * @param rhs VectorD of same size.
     * @return VectorD of same size.
     */
    VectorD operator -(const VectorD &rhs) const;
    /**
     * Addition.
     * @param rhs VectorD of same size.
     * @return VectorD of same size.
     */
    VectorD operator +(const VectorD &rhs) const;

    // Vector/double operations
    /**
     * Multiplication of all coefficients by a scalar.
     * @param rhs Scalar.
     * @return Vector.
     */
    VectorD operator*(double rhs) const;
    /**
     * Addition of all coefficients with a scalar.
     * @param rhs Scalar.
     * @return Vector.
     */
    VectorD operator+(double rhs) const;
    /**
     * Subtraction of all coefficient with a scalar.
     * @param rhs Scalar.
     * @return Vector.
     */
    VectorD operator-(double rhs) const;
    /**
     * Division of all coefficients by a scalar.
     * @param rhs Scalar.
     * @return Vector.
     */
    VectorD operator/(double rhs) const;

    /**
     * Norm 2.
     * @return double
     */
    double norm() const;

    /**
     * Conversion to a matrix of one column.
     * @return MatrixD typed vector column.
     */
    MatrixD toMatrix();

    /**
     * Direct accessor in memory with the specified index value n.
     * @param n
     * @return double
     */
    double& operator [](size_t n);
    double& at(size_t n); ///< Same as std::vector.

    /**
     * Direct accessor in memory with the specified index value n.
     * @param n
     * @return double
     */
    const double& operator [](size_t n) const;
    const double& at(size_t n) const; ///< Same as std::vector.

    /**
     * Concatenation of two vectors.
     * @param V1
     * @param V2
     * @return VectorD of size V1.size() + V2.size()
     */
    static VectorD concatenate(const VectorD& V1, const VectorD& V2);

    const std::vector<double>& toStdVector() const; ///< Conversion to std::vector.

    /**
     * Return a formatted characters string to display vector's values.
     *
     * @return std::string
     */
    std::string toString() const;

private:
    std::vector<double> m_vector; ///< An array of double to represent the vector in memory.
};

} /* namespace NumA */

#endif /* VECTOR_D_H */
