/*
 * VectorComplexD.h
 *
 *  Created on: Jan 16, 2016
 *      Author: Pawel Sznajder (IPNO)
 */

#ifndef VECTORCOMPLEXD_H_
#define VECTORCOMPLEXD_H_

#include <complex>
#include <vector>

namespace NumA {

/**
 * @class VectorComplexD
 *
 * @brief Vector of complex numbers of undefined size.
 *
 * This class represents a vector of complex double precision numbers. The size of this vector should be defined by the user.
 */
class VectorComplexD {

public:

    /**
     * Default constructor.
     */
    VectorComplexD();

    /**
     * Assignment constructor.
     * Elements will be initialized with `std::complex<double>(0., 0.).`
     * @param n Number of elements to be set.
     */
    VectorComplexD(const unsigned int n);

    /**
     * Copy constructor.
     * @param v Object to be copied.
     */
    VectorComplexD(const VectorComplexD& v);

    /**
     * Destructor.
     */
    virtual ~VectorComplexD();

    /**
     * Get magnitude of vector.
     * Return: \f$\sqrt{f \cdot f^{*}}\f$.
     */
    double Mag() const;

    /**
     * Get conjunction of vector.
     * Return: \f$f^{*}\f$.
     */
    VectorComplexD Conjunct() const;

    /**
     * Scalar product of this and other vector.
     * Return: \f$f \cdot g\f$.
     * @param v Other vector.
     */
    std::complex<double> Dot(const VectorComplexD& v) const;

    /**
     * Clear vector (set `std::complex<double>(0., 0.)` to all elements).
     */
    void Clear();

    /**
     * Operator =.
     */
     void operator=(const VectorComplexD& rhs);

    /**
     * Operator +.
     */
     VectorComplexD operator+(const VectorComplexD& rhs);

    /**
     * Operator -.
     */
     VectorComplexD operator-(const VectorComplexD& rhs);

    /**
     * Operator * by real number (rhs).
     */
     VectorComplexD operator*(const double& rhs);

    /**
     * Operator * by real number (lhs).
     */
    friend VectorComplexD operator*(const double& lhs, VectorComplexD& rhs);

    /**
     * Operator * by complex number (rhs).
     */
     VectorComplexD operator*(const std::complex<double>& rhs);

    /**
     * Operator * by complex number (lhs).
     */
    friend VectorComplexD operator*(const std::complex<double>& lhs,
            VectorComplexD& rhs);

    /**
     * Get number of elements.
     */
    unsigned int GetNElements() const;

    /**
     * Set element.
     * @param i Index of element to be set.
     * @param value Value to be set.
     */
    void SetElement(const unsigned int i, const std::complex<double>& value);

    /**
     * Get element.
     * @param i Index of element to be retrieved.
     */
    std::complex<double> GetElement(unsigned int i) const;

private:

    /**
     * Compare number of elements of this and other vector. Throw `std::runtime_error` if the numbers are not equal.
     * @param v Other vector.
     */
    void CompareNElements(const VectorComplexD& v) const;

    /**
     * Number of elements.
     */
    unsigned int m_n;

    /**
     * Elements.
     */
    std::vector<std::complex<double> > m_elements;
};

} /* namespace NumA */

#endif /* VECTORCOMPLEXD_H_ */
