#ifndef NEWTON_H
#define NEWTON_H

/**
 * @file Newton.h
 * @author Bryan BERTHOU (SPhN / CEA Saclay)
 * @date 29 January 2016
 * @version 1.0
 */

#include <ElementaryUtils/logger/CustomException.h>
#include <stddef.h>
#include <cmath>
#include <iostream>
#include <vector>

namespace NumA {

/**
 * @class Newton
 *
 * @brief The code supports on the example of the python library : scipy.
 * Find a zero using the Newton-Raphson or secant method.
 *
 * Find a zero of the function `func` given a nearby starting point `x0`.
 * The Newton-Raphson method is used if the derivative `fprime` of `func`
 * is provided, otherwise the secant method is used.
 */
class Newton {
public:
    //TODO add relative tolerance ; return status message when max iteration reached and no convergence
    template<class OBJ, typename FUNC, typename DERIV>
    double solve(OBJ* object, FUNC function,
            std::vector<double> &funcParameters, double x0,
            DERIV derivative = 0, const double absTolerance = 1.48e-8,
            const size_t maxIteration = 50) {

        if (absTolerance <= 0.) {
            throw ElemUtils::CustomException("Newton", __func__,
                    "Tolerance too small");
        }
        if (maxIteration < 1) {
            throw ElemUtils::CustomException("Newton", __func__,
                    "Number of maximum iteration must be greater than 0");
        }

        std::vector<double> variable(1, 0.);
        double x = 0.;
        bool noConvergence = true;

        if (derivative != 0) {
            // Newton-Rapheson method
            double result_deriv = 0., result = 0.;

            for (size_t i = 0; (i < maxIteration) && noConvergence; i++) {
                result_deriv = ((*object).*derivative)(x0, funcParameters);
                if (result_deriv == 0.) {
                    std::cerr
                            << "[WARNING] (Newton::solve) Derivative was zero."
                            << std::endl;
                    return x0;
                }

                result = ((*object).*function)(x0, funcParameters);

                // Newton step
                x = x0 - result / result_deriv;
                noConvergence = fabs(x - x0) > absTolerance;

                x0 = x;
            }

            return x;

        } else {
            // Secant method
            double x1 = 0.;

            if (x0 >= 0.) {
                x1 = x0 * (1 + 1e-4) + 1e-4;
            } else {
                x1 = x0 * (1 + 1e-4) - 1e-4;
            }

            double v0 = ((*object).*function)(x0, funcParameters);
            double v1 = ((*object).*function)(x1, funcParameters);

            for (size_t i = 0; (i < maxIteration) && noConvergence; i++) {
                if (v1 == v0) {

                    if (x1 != x0) {
                        std::cerr << "[WARNING] (Newton::solve) Tolerance of "
                                << (x1 - x0) << "reached" << std::endl;
                    }

                    return (x1 + x0) / 2.0;
                } else {
                    x = x1 - v1 * (x1 - x0) / (v1 - v0);
                }

                noConvergence = fabs(x - x1) > absTolerance;

                x0 = x1;
                v0 = v1;
                x1 = x;
                v1 = ((*object).*function)(x1, funcParameters);
            }

            return x;
        }
    }
};

} /* namespace NumA */

#endif /* NEWTON_H */
