/**
 * @file NewtonMD.h
 * @author Nabil CHOUIKA (SPhN / CEA Saclay)
 * @date Feb 5 2016
 * @version 1.0
 */

#ifndef NEWTONMD_H_
#define NEWTONMD_H_

#include <stddef.h>

#include "../linear_algebra/LinAlgUtils.h"
#include "../linear_algebra/matrix/MatrixD.h"
#include "../linear_algebra/vector/VectorD.h"

namespace NumA {

/**
 * @class NewtonMD
 *
 * @brief Newton method generalized to any dimension N:
 * Solves the system G(X) = 0,
 * where G is a regular but non-linear application \f$ R^N \rightarrow R^N \f$ .
 *
 * The class can be used either through templates with the solve method
 * providing G and the Jacobian matrix functions pointers, or manually
 * through instantiating an object NewtonMD and iterating the method iterate
 * by giving G(X) and the Jacobian matrix at each step, to avoid using functions.
 *
 * For linear G (matrices), use the LinearSystem class.
 */
class NewtonMD {
public:
    NewtonMD(int N = 1, LinAlgUtils::Method linMethod = LinAlgUtils::ColQR);
    virtual ~NewtonMD();

    int getN() const;
    void setN(int n);
    LinAlgUtils::Method getLinMethod() const;
    void setLinMethod(LinAlgUtils::Method linMethod);

    template<class OBJ, typename FUNC, typename JACOB>
    VectorD solve(OBJ* object, FUNC G,
            const std::vector<double> &funcParameters, VectorD X0,
            JACOB J_G = 0, const double absTolerance = 1.e-4,
            const double relTolerance = 1.e-4, const size_t maxIteration = 50) {

        bool noConvergence = true;
        double absDiff = 0, relDiff = 0.;
        VectorD X(X0.size());

        for (unsigned int n = 0; n < maxIteration && noConvergence; n++) {

            VectorD G_X0 = ((*object).*G)(X0, funcParameters);
            MatrixD J_G_X0 = ((*object).*J_G)(X0, funcParameters);

            X = iterate(X0, G_X0, J_G_X0);

            absDiff = (X - X0).norm();
            relDiff = absDiff / (X0.norm() + 1.e-16);
            noConvergence = absDiff > absTolerance || relDiff > relTolerance;

            X0 = X;
        }

        return X;
    }

    /** Produces a new iteration by solving the linear system
     * J^G(X0) * (X0 - X1) = G(X0)
     * and returns X1
     *
     * @param X0 : VectorD of size N
     * @param G_X0 : VectorD of size N
     * @param J_G_X0 : MatrixD of size NxN
     * @return X1 : New solution after the new iteration.
     */
    VectorD iterate(const VectorD & X0, const VectorD & G_X0,
            const MatrixD & J_G_X0);

private:
    int m_N; ///< Dimension of the problem
    LinAlgUtils::Method m_linMethod; ///< Method used for solving linear systems
};

} /* namespace NumA */

#endif /* NEWTONMD_H_ */
