#ifndef FUNCTOR_UTILS_H
#define FUNCTOR_UTILS_H

/**
 * @file FunctorUtils.h
 * @author Bryan BERTHOU (SPhN / CEA Saclay)
 * @date October 07, 2016
 * @version 1.0
 */

#include "../functor/multi_dimension/FunctorMD.h"
#include "../functor/one_dimension/Functor1D.h"

namespace NumA {

/**
 * @class FunctorUtils
 *
 * @brief Utilities for Functors.
 */
class FunctorUtils {
public:
    /**
     * Defines a new functor with the given one-dimensional function.
     *
     * See FunctionType1D documentation for an example.
     * @param object Pointer to the object instantiating a class where the function is defined as a method.
     * @param function Reference to the function defining the functor. Should have the following signature:
     * ```cpp
     * double MyClass::myFunction(double variable, std::vector<double>& parameters);
     * ```
     * @return Functor pointer.
     */
    template<typename PointerToObj, typename PointerToMemFn>
    static Functor1D<PointerToObj, PointerToMemFn>* newFunctor1D(
            PointerToObj* object, PointerToMemFn function) {
        return new Functor1D<PointerToObj, PointerToMemFn>(object, function);
    }

    /**
     * Defines a new functor with the given multi-dimensional function.
     *
     * See FunctionTypeMD documentation for an example.
     * @param object Pointer to the object instantiating a class where the function is defined as a method.
     * @param function Reference to the function defining the functor. Should have the signature:
     * ```cpp
     * double MyClass::myFunction(NumA::VectorD& variables, std::vector<double>& parameters);
     * ```
     * @return Functor pointer.
     */
    template<typename PointerToObj, typename PointerToMemFn>
    static FunctorMD<PointerToObj, PointerToMemFn>* newFunctorMD(
            PointerToObj* object, PointerToMemFn function) {
        return new FunctorMD<PointerToObj, PointerToMemFn>(object, function);
    }
};

} /* namespace NumA */

#endif /* FUNCTOR_UTILS_H */
