#ifndef INTERVAL_H
#define INTERVAL_H

/**
 * @file Interval.h
 * @author Bryan BERTHOU (SPhN / CEA Saclay)
 * @date July 23, 2014
 * @version 1.0
 */

#include <ElementaryUtils/logger/CustomException.h>
#include <stddef.h>
#include <cmath>
#include <vector>

namespace NumA {

/**
 * @class Interval
 *
 * @brief Class defining an interval (with given bounds and step).
 */
template<typename T>
class Interval {
public:

    /**
     * Type of interval.
     */
    enum StepMode {
        NORMAL = 0, ///< Linear interval.
        LOG = 1 ///< Logarithmic interval.
    };

    /**
     * Constructor.
     * @param lowerBound Lower bound of the interval.
     * @param upperBound Upper bound of the interval.
     * @param step Value of the step.
     * @param stepMode Type of interval.
     */
    Interval(T lowerBound, T upperBound, T step = 1,
            Interval::StepMode stepMode = NORMAL) :
            m_lowerBound(lowerBound), m_upperBound(upperBound), m_step(step), m_stepMode(
                    stepMode) {
    }

    /**
     *
     * @return Nodes of interval.
     */
    std::vector<T> computeSteps() const {
        std::vector<T> steps;

        switch (m_stepMode) {
        case NORMAL: {

            T i = m_lowerBound;
            while (i < m_upperBound) {
                steps.push_back(i);
                i += m_step;
            }

            steps.push_back(m_upperBound);

            break;
        }
        case LOG: {
//TODO implement
            break;
        }
        default:
            throw ElemUtils::CustomException("Interval", __func__,
                    "Unknown step mode");
            break;
        }

        return steps;
    }

    /**
     * Static function that computes the nodes of an interval for a chosen number of nodes.
     * @param lowerBound Lower bound of the interval.
     * @param upperBound Upper bound of the interval.
     * @param num Number of nodes.
     * @param stepMode Type of interval.
     * @return Nodes.
     */
    static std::vector<T> computeNodes(T lowerBound, T upperBound, size_t num =
            10, Interval::StepMode stepMode = NORMAL) {
        T node = lowerBound;
        T step;
        std::vector<T> nodes(num, node);
        switch (stepMode) {
        case NORMAL:
            step = (upperBound - lowerBound) / (num - 1);
            for (size_t i = 1; i < num; i++) {
                node += step;
                nodes[i] = node;
            }
            break;
        case LOG:
            step = pow((upperBound / lowerBound), 1 / (num - 1));
            for (size_t i = 1; i < num; i++) {
                node *= step;
                nodes[i] = node;
            }
            break;
        default:
            throw ElemUtils::CustomException("Interval", __func__,
                    "Unknown step mode");
            break;
        }
        return nodes;
    }

    /**
     *
     * @return Lower bound of the interval.
     */
    T getLowerBound() const {
        return m_lowerBound;
    }

    /**
     *
     * @param lowerBound Lower bound of the interval.
     */
    void setLowerBound(T lowerBound) {
        this->m_lowerBound = lowerBound;
    }

    /**
     *
     * @return Step.
     */
    T getStep() const {
        return m_step;
    }

    /**
     *
     * @param step Value of the step.
     */
    void setStep(T step) {
        this->m_step = step;
    }

    /**
     *
     * @return Upper bound of the interval.
     */
    T getUpperBound() const {
        return m_upperBound;
    }

    /**
     *
     * @param upperBound Upper bound of the interval.
     */
    void setUpperBound(T upperBound) {
        this->m_upperBound = upperBound;
    }

private:
    T m_lowerBound; ///< Lower bound of the interval.
    T m_upperBound; ///< Upper bound of the interval.
    T m_step; ///< Step.

    Interval::StepMode m_stepMode; ///< Type of interval.
};

} /* namespace NumA */

#endif /* INTERVAL_H */
