#include "../../../include/NumA/gsl/ExtrapolationTable.h"

#include <cmath>

#include "../../../include/NumA/gsl/GSLConstants.h"
#include "../../../include/NumA/gsl/GSLUtils.h"

ExtrapolationTable::ExtrapolationTable() :
        m_n(0), m_nres(0), m_rlist2(std::vector<double>(52)), m_res3la(
                std::vector<double>(3)) {
}

ExtrapolationTable::~ExtrapolationTable() {
}

void ExtrapolationTable::append_table(double y) {
    size_t n = m_n;
    m_rlist2[n] = y;
    m_n++;
}

void ExtrapolationTable::qelg(double& result, double& abserr) {
    std::vector<double> epstab = m_rlist2;
    std::vector<double> res3la = res3la;
    const size_t n = m_n - 1;

    const double current = epstab[n];

    double absolute = GSL_DBL_MAX;
    double relative = 5. * GSL_DBL_EPSILON * fabs(current);

    const size_t newelm = n / 2;
    const size_t n_orig = n;
    size_t n_final = n;
    size_t i;

    const size_t nres_orig = m_nres;

    result = current;
    abserr = GSL_DBL_MAX;

    if (n < 2) {
        abserr = GSLUtils::GSL_MAX_DBL(absolute, relative);
        return;
    }

    epstab[n + 2] = epstab[n];
    epstab[n] = GSL_DBL_MAX;

    for (i = 0; i < newelm; i++) {
        double res = epstab[n - 2 * i + 2];
        double e0 = epstab[n - 2 * i - 2];
        double e1 = epstab[n - 2 * i - 1];
        double e2 = res;

        double e1abs = fabs(e1);
        double delta2 = e2 - e1;
        double err2 = fabs(delta2);
        double tol2 = GSLUtils::GSL_MAX_DBL(fabs(e2), e1abs) * GSL_DBL_EPSILON;
        double delta3 = e1 - e0;
        double err3 = fabs(delta3);
        double tol3 = GSLUtils::GSL_MAX_DBL(e1abs, fabs(e0)) * GSL_DBL_EPSILON;

        double e3 = 0., delta1 = 0., err1 = 0., tol1 = 0., ss = 0.;

        if (err2 <= tol2 && err3 <= tol3) {
            /* If e0, e1 and e2 are equal to within machine accuracy,
             convergence is assumed.  */

            result = res;
            absolute = err2 + err3;
            relative = 5. * GSL_DBL_EPSILON * fabs(res);
            abserr = GSLUtils::GSL_MAX_DBL(absolute, relative);
            return;
        }

        e3 = epstab[n - 2 * i];
        epstab[n - 2 * i] = e1;
        delta1 = e1 - e3;
        err1 = fabs(delta1);
        tol1 = GSLUtils::GSL_MAX_DBL(e1abs, fabs(e3)) * GSL_DBL_EPSILON;

        /* If two elements are very close to each other, omit a part of
         the table by adjusting the value of n */

        if (err1 <= tol1 || err2 <= tol2 || err3 <= tol3) {
            n_final = 2 * i;
            break;
        }

        ss = ((1. / delta1) + (1. / delta2)) - (1. / delta3);

        /* Test to detect irregular behaviour in the table, and
         eventually omit a part of the table by adjusting the value of
         n. */

        if (fabs(ss * e1) <= 0.0001) {
            n_final = 2 * i;
            break;
        }

        /* Compute a new element and eventually adjust the value of
         result. */

        res = e1 + (1. / ss);
        epstab[n - 2 * i] = res;

        {
            const double error = err2 + fabs(res - e2) + err3;

            if (error <= abserr) {
                abserr = error;
                result = res;
            }
        }
    }

    /* Shift the table */

    {
        const size_t limexp = 50 - 1;

        if (n_final == limexp) {
            n_final = 2 * (limexp / 2);
        }
    }

    if (n_orig % 2 == 1) {
        for (i = 0; i <= newelm; i++) {
            epstab[1 + i * 2] = epstab[i * 2 + 3];
        }
    } else {
        for (i = 0; i <= newelm; i++) {
            epstab[i * 2] = epstab[i * 2 + 2];
        }
    }

    if (n_orig != n_final) {
        for (i = 0; i <= n_final; i++) {
            epstab[i] = epstab[n_orig - n_final + i];
        }
    }

    m_n = n_final + 1;

    if (nres_orig < 3) {
        res3la[nres_orig] = result;
        abserr = GSL_DBL_MAX;
    } else { /* Compute error estimate */
        abserr = (fabs(result - res3la[2]) + fabs(result - res3la[1])
                + fabs(result - res3la[0]));

        res3la[0] = res3la[1];
        res3la[1] = res3la[2];
        res3la[2] = result;
    }

    /* In QUADPACK the variable table->nres is incremented at the top of
     qelg, so it increases on every call. This leads to the array
     res3la being accessed when its elements are still undefined, so I
     have moved the update to this point so that its value more
     useful. */

    m_nres = nres_orig + 1;

    abserr = GSLUtils::GSL_MAX_DBL(abserr, 5. * GSL_DBL_EPSILON * fabs(result));
}
