#include <ElementaryUtils/logger/CustomException.h>
#include <ElementaryUtils/string_utils/Formatter.h>
#include <include/NumA/functor/one_dimension/FunctionType1D.h>
#include <include/NumA/integration/one_dimension/TrapezoidalLogIntegrator1D.h>
#include <stddef.h>
#include <cmath>

namespace NumA {

TrapezoidalLogIntegrator1D::TrapezoidalLogIntegrator1D(unsigned int N) :
		QuadratureIntegrator1D(N) {
}

TrapezoidalLogIntegrator1D::TrapezoidalLogIntegrator1D(
		const TrapezoidalLogIntegrator1D &other) :
		QuadratureIntegrator1D(other) {
}

TrapezoidalLogIntegrator1D::~TrapezoidalLogIntegrator1D() {
}

TrapezoidalLogIntegrator1D* TrapezoidalLogIntegrator1D::clone() const {
	return new TrapezoidalLogIntegrator1D(*this);
}

double TrapezoidalLogIntegrator1D::integrate(FunctionType1D* pFunction,
		double a, double b, std::vector<double> &parameters) {

	//check if valid
	if (a <= 0. || b <= 0.) {
		throw ElemUtils::CustomException("TrapezoidalLogIntegrator1D", __func__,
				ElemUtils::Formatter()
						<< "This integration routine can be used only for > 0 range, here: "
						<< a << ", " << b);
	}

	//result
	double result = 0.;

	//log range and step
	double logA = log(a);
	double logB = log(b);
	double logStep = (logB - logA) / double(m_N - 1);

	double argumentA = a;
	double argumentB;

	double valueA = ((*pFunction)(a, parameters));
	double valueB;

	// Entering the loop
	for (size_t i = 1; i < m_N; i++) {

		//actual
		argumentB = exp(logA + i * logStep);

		//calculate
		valueB = ((*pFunction)(argumentB, parameters));

		//add
		result += (argumentB - argumentA) * (valueA + valueB);

		//save previous
		argumentA = argumentB;
		valueA = valueB;
	}

	//return
	return 0.5 * result;
}

void TrapezoidalLogIntegrator1D::configure(
		const ElemUtils::Parameters &parameters) {
	QuadratureIntegrator1D::configure(parameters);
}

void TrapezoidalLogIntegrator1D::setN(unsigned int n) {
	m_N = n;
}

} /* namespace NumA */
