#include "../../../../../include/NumA/integration/one_dimension/gauss_kronrod_adaptive/Workspace.h"

namespace NumA {

Workspace::Workspace(double limit) :
        m_limit(limit), m_size(0), m_nrmax(0), m_i(0), m_maximum_level(0) {
    m_alist = std::vector<double>(limit);
    m_blist = std::vector<double>(limit);
    m_rlist = std::vector<double>(limit);
    m_elist = std::vector<double>(limit);
    m_order = std::vector<size_t>(limit);
    m_level = std::vector<size_t>(limit);
}

Workspace::~Workspace() {
}

void Workspace::initialise(double a, double b) {
    m_alist[0] = a;
    m_blist[0] = b;
    m_rlist[0] = 0.0;
    m_elist[0] = 0.0;
    m_order[0] = 0;
    m_level[0] = 0;
}

void Workspace::set_initial_result(double result, double error) {
    m_size = 1;
    m_rlist[0] = result;
    m_elist[0] = error;
}

void Workspace::retrieve(double& a, double& b, double& r, double& e) {
    a = m_alist[m_i];
    b = m_blist[m_i];
    r = m_rlist[m_i];
    e = m_elist[m_i];
}

void Workspace::update(double a1, double b1, double area1, double error1,
        double a2, double b2, double area2, double error2) {
    std::vector<double> alist = m_alist;
    std::vector<double> blist = m_blist;
    std::vector<double> rlist = m_rlist;
    std::vector<double> elist = m_elist;
    std::vector<size_t> level = m_level;

    const size_t i_max = m_i;
    const size_t i_new = m_size;

    const size_t new_level = m_level[i_max] + 1;

    /* append the newly-created intervals to the list */

    if (error2 > error1) {
        alist[i_max] = a2; /* blist[maxerr] is already == b2 */
        rlist[i_max] = area2;
        elist[i_max] = error2;
        level[i_max] = new_level;

        alist[i_new] = a1;
        blist[i_new] = b1;
        rlist[i_new] = area1;
        elist[i_new] = error1;
        level[i_new] = new_level;
    } else {
        blist[i_max] = b1; /* alist[maxerr] is already == a1 */
        rlist[i_max] = area1;
        elist[i_max] = error1;
        level[i_max] = new_level;

        alist[i_new] = a2;
        blist[i_new] = b2;
        rlist[i_new] = area2;
        elist[i_new] = error2;
        level[i_new] = new_level;
    }

    m_size++;

    if (new_level > m_maximum_level) {
        m_maximum_level = new_level;
    }

    qpsrt();
}

void Workspace::qpsrt() {
    const size_t last = m_size - 1;
    const size_t limit = m_limit;

    std::vector<double> elist = m_elist;
    std::vector<size_t> order = m_order;

    double errmax;
    double errmin;
    int i, k, top;

    size_t i_nrmax = m_nrmax;
    size_t i_maxerr = order[i_nrmax];

    /* Check whether the list contains more than two error estimates */

    if (last < 2) {
        order[0] = 0;
        order[1] = 1;
        m_i = i_maxerr;
        return;
    }

    errmax = elist[i_maxerr];

    /* This part of the routine is only executed if, due to a difficult
     integrand, subdivision increased the error estimate. In the normal
     case the insert procedure should start after the nrmax-th largest
     error estimate. */

    while (i_nrmax > 0 && errmax > elist[order[i_nrmax - 1]]) {
        order[i_nrmax] = order[i_nrmax - 1];
        i_nrmax--;
    }

    /* Compute the number of elements in the list to be maintained in
     descending order. This number depends on the number of
     subdivisions still allowed. */

    if (last < (limit / 2 + 2)) {
        top = last;
    } else {
        top = limit - last + 1;
    }

    /* Insert errmax by traversing the list top-down, starting
     comparison from the element elist(order(i_nrmax+1)). */

    i = i_nrmax + 1;

    /* The order of the tests in the following line is important to
     prevent a segmentation fault */

    while (i < top && errmax < elist[order[i]]) {
        order[i - 1] = order[i];
        i++;
    }

    order[i - 1] = i_maxerr;

    /* Insert errmin by traversing the list bottom-up */

    errmin = elist[last];

    k = top - 1;

    while (k > i - 2 && errmin >= elist[order[k]]) {
        order[k + 1] = order[k];
        k--;
    }

    order[k + 1] = last;

    /* Set i_max and e_max */

    i_maxerr = order[i_nrmax];

    m_i = i_maxerr;
    m_nrmax = i_nrmax;
}

/* The smallest interval has the largest error.  Before bisecting
 * decrease the sum of the errors over the larger intervals
 * (error_over_large_intervals) and perform extrapolation. */
int Workspace::increase_nrmax() {
    int k;
    int id = m_nrmax;
    int jupbnd;

    const std::vector<size_t> level = m_level;
    const std::vector<size_t> order = m_order;

    size_t limit = m_limit;
    size_t last = m_size - 1;

    if (last > (1 + limit / 2)) {
        jupbnd = limit + 1 - last;
    } else {
        jupbnd = last;
    }

    for (k = id; k <= jupbnd; k++) {
        size_t i_max = order[m_nrmax];

        m_i = i_max;

        if (level[i_max] < m_maximum_level) {
            return 1;
        }

        m_nrmax++;

    }

    return 0;
}

void Workspace::reset_nrmax() {
    m_nrmax = 0;
    m_i = m_order[0];
}

double Workspace::sum_results() {
    const std::vector<double> rlist = m_rlist;
    const size_t n = m_size;

    size_t k;
    double result_sum = 0;

    for (k = 0; k < n; k++) {
        result_sum += rlist[k];
    }

    return result_sum;
}

int Workspace::large_interval() {
    size_t i = m_i;
    const std::vector<size_t> level = m_level;

    if (level[i] < m_maximum_level) {
        return 1;
    } else {
        return 0;
    }
}

} // namespace NumA
