#include "../../../include/NumA/interpolation/Chebyshev.h"

#include <cmath>
#include <stdexcept>
#include <sstream>

namespace NumA {

Chebyshev::Chebyshev(unsigned int N) :
        m_N(N) {
    initRoots();
    initValuesToCoeffsMatrix();
}

Chebyshev::~Chebyshev() {
    // TODO Auto-generated destructor stub
}

unsigned int Chebyshev::getN() const {
    return m_N;
}

void Chebyshev::setN(unsigned int n) {
    m_N = n;
    initRoots();
    initValuesToCoeffsMatrix();
}

const NumA::VectorD& Chebyshev::getRoots() const {
    return m_roots;
}

const NumA::MatrixD& Chebyshev::getValuesToCoeffsMatrix() const {
    return m_valuesToCoeffsMatrix;
}

const NumA::MatrixD& Chebyshev::getCoeffsToValuesMatrix() const {
    return m_coeffsToValuesMatrix;
}

double Chebyshev::T(unsigned int i, double x) {
    return cos(i * acos(x));
}

void Chebyshev::initRoots() {
    m_roots.assign(m_N, 0.);
    for (int k = 0; k < m_N; k++) {
        m_roots.at(k) = cos((m_N - k - 0.5) * M_PI / m_N);
    }
    m_coeffsToValuesMatrix = getCoeffsToValuesMatrix(m_roots);
}

void Chebyshev::initValuesToCoeffsMatrix() {
    m_valuesToCoeffsMatrix.assign(m_N, m_N, 0.);

    for (int i = 0; i < m_N; i++) {
        for (int k = 0; k < m_N; k++) {
            m_valuesToCoeffsMatrix.at(i, k) = cos(
                    i * (m_N - k - 0.5) * M_PI / m_N);
        }
    }
    m_valuesToCoeffsMatrix = m_valuesToCoeffsMatrix * (2. / m_N);
}

NumA::MatrixD Chebyshev::getCoeffsToValuesMatrix(
        const NumA::VectorD& points) const {
    NumA::MatrixD result(points.size(), m_N);

    for (unsigned int i = 0; i < points.size(); i++) {
        result.setLine(i, getCoeffsToValueVector(points.at(i)));
    }

    return result;
}

NumA::VectorD Chebyshev::getCoeffsToValueVector(double x) const {
    if (x > 1. || x < -1.) {
        std::stringstream formatter;
        formatter << "[Chebyshev::getCoeffsToValueVector] Point " << x
                << " out of bounds [-1,1] of Chebyshev expansion.";
        throw std::runtime_error(formatter.str());
    }
    NumA::VectorD result(m_N, 0.);
    result.at(0) = 0.5;
    for (unsigned int i = 1; i < m_N; i++) {
        result.at(i) = T(i, x);
    }
    return result;
}

NumA::MatrixD Chebyshev::getValuesToValuesMatrix(
        const NumA::VectorD& points) const {
    return getCoeffsToValuesMatrix(points) * getValuesToCoeffsMatrix();
}

NumA::VectorD Chebyshev::getValuesToValueVector(double x) const {
    return getValuesToCoeffsMatrix().transpose() * getCoeffsToValueVector(x);
}

double Chebyshev::evaluateExpansion(const NumA::VectorD& coeffs, double x) const {
    if (coeffs.size() != m_N) {
        std::stringstream formatter;
        formatter << "[Chebyshev::evaluateExpansion] Number of coefficients "
                << coeffs.size()
                << " does not match the order of the expansion " << m_N;
        throw std::runtime_error(formatter.str());
    }
    return getCoeffsToValueVector(x) * coeffs;
}

} /* namespace NumA */
