#include "../../../include/NumA/linear_algebra/LinAlgUtils.h"

#include <algorithm>
#include <cmath>
#include <ctime>
#include <Eigen/Core>
#include <Eigen/LU>
#include <Eigen/QR>
#include <limits>
#include <string>
#include <vector>
#include <ElementaryUtils/logger/CustomException.h>
#include <ElementaryUtils/string_utils/Formatter.h>

#include "../../../include/NumA/linear_algebra/eigen/EigenUtils.h"

namespace NumA {

VectorD LinAlgUtils::solve(const MatrixD & A, const VectorD & Y,
        Method method) {
    if (A.cols() != A.rows()) {
        ElemUtils::Formatter formatter;
        formatter << "Matrix column size = " << A.cols()
                << " does not match matrix row size = " << A.rows();
        throw ElemUtils::CustomException("LinAlgUtils", __func__,
                formatter.str());
    }
    if (Y.size() != A.rows()) {
        ElemUtils::Formatter formatter;
        formatter << "Vector's size = " << Y.size()
                << " does not match matrix size = " << A.rows();
        throw ElemUtils::CustomException("LinAlgUtils", __func__,
                formatter.str());
    }

    // Converting matrix and vector to Eigen
    Eigen::MatrixXd EigenA = EigenUtils::convertToEigen(A);
    Eigen::VectorXd EigenY = EigenUtils::convertToEigen(Y);

    // Solving the system
    Eigen::VectorXd EigenX;
    switch (method) {
    case FullQR:
        EigenX = EigenA.fullPivHouseholderQr().solve(EigenY);
        break;
    case QR:
        EigenX = EigenA.householderQr().solve(EigenY);
        break;
    case PartialLU:
        EigenX = EigenA.partialPivLu().solve(EigenY);
        break;
    case FullLU:
        EigenX = EigenA.fullPivLu().solve(EigenY);
        break;
    default:
        EigenX = EigenA.colPivHouseholderQr().solve(EigenY);
        break;
    }

    // Converting to VectorD
    VectorD X = EigenUtils::convertToVectorD(EigenX);

    return X;
}

unsigned int LinAlgUtils::rank(const MatrixD& A, Method method) {
    // Converting matrix and vector to Eigen
    Eigen::MatrixXd EigenA = EigenUtils::convertToEigen(A);

    unsigned int rank;
    switch (method) {
    case FullQR:
        rank = EigenA.fullPivHouseholderQr().rank();
        break;
    case FullLU:
        rank = EigenA.fullPivLu().rank();
        break;
    default:
        rank = EigenA.colPivHouseholderQr().rank();
        break;
    }

    return rank;
}

} /* namespace NumA */
