#include "../../../../include/NumA/linear_algebra/least_squares/LSMRSolver.h"

#include <algorithm>
#include <cmath>
#include <ctime>
#include <limits>
#include <ElementaryUtils/string_utils/Formatter.h>

#include "../../../../include/NumA/linear_algebra/eigen/EigenUtils.h"

namespace NumA {

LSMRSolver::LSMRSolver(double damp, double atol, double btol, double conlim,
        size_t maxiter, ElemUtils::Formatter* output) :
        m_damp(damp), m_atol(atol), m_btol(btol), m_conlim(conlim), m_maxiter(
                maxiter), m_output(output), m_istop(0), m_itn(0), m_normr(0.), m_normar(
                0.), m_normA(0.), m_condA(0.), m_normx(0.), m_pfreq(20) {
    m_msg.assign(8, "");
    m_msg[0] = "The exact solution is x = 0.";
    m_msg[1] = "Ax - b is small enough, given atol, btol.";
    m_msg[2] = "The least-squares solution is good enough, given atol.";
    m_msg[3] = "The estimate of cond(Abar) has exceeded conlim.";
    m_msg[4] = "Ax - b is small enough for this machine.";
    m_msg[5] = "The least-squares solution is good enough for this machine.";
    m_msg[6] = "Cond(Abar) seems to be too large for this machine.";
    m_msg[7] = "The iteration limit has been reached.";
    m_hdg1 = "  itn x(1) norm r norm A'r";
    m_hdg2 = " compatible LS norm A cond A";
}

LSMRSolver::~LSMRSolver() {
    // TODO Auto-generated destructor stub
}

VectorD LSMRSolver::solve(const MatrixD& A, const VectorD& B) {

    // Converting matrix and vector to Eigen
    Eigen::MatrixXd EigenA = EigenUtils::convertToEigen(A);
    Eigen::VectorXd EigenB = EigenUtils::convertToEigen(B);

    size_t m = A.rows();
    size_t n = A.cols();
//    MatrixD At = A.transpose();
    Eigen::MatrixXd EigenAt = EigenA.transpose();

    // stores the num of singular values
    size_t minDim = std::min(m, n);

    if (m_maxiter <= 0)
        m_maxiter = minDim;

    if (m_output) {
        *m_output << " ";
        *m_output << "LSMR Least-squares solution of Ax = b\n";
        *m_output << "The matrix A has " << m << " rows and " << n << " cols\n";
        *m_output << "damp = " << m_damp << "\n";
        *m_output << "atol = " << m_atol << "  conlim = " << m_conlim << "\n";
        *m_output << "btol = " << m_btol << "  maxiter = " << m_maxiter << "\n";
    }

//    VectorD u(Y);
    Eigen::VectorXd u(EigenB);
    double beta = u.norm();

//    VectorD v(n);
    Eigen::VectorXd v(n);
    v.setZero();
    double alpha = 0.;

    if (beta > 0) {
        u = u / beta;
//        v = At * u;
        v = EigenAt * u;
        alpha = v.norm();
    }

    if (alpha > 0) {
        v = v / alpha;
    }

    // Initialize variables for 1st iteration.

    unsigned int pcount = 0; // print counter
    m_itn = 0;
    double zetabar = alpha * beta;
    double alphabar = alpha;
    double rho = 1;
    double rhobar = 1;
    double cbar = 1;
    double sbar = 0;

//    VectorD h(v);
//    VectorD hbar(n);
//    VectorD x(n);
    Eigen::VectorXd h(v);
    Eigen::VectorXd hbar(n);
    hbar.setZero();
    Eigen::VectorXd x(n);
    x.setZero();

    // Initialize variables for estimation of ||r||.

    double betadd = beta;
    double betad = 0.;
    double rhodold = 1.;
    double tautildeold = 0.;
    double thetatilde = 0.;
    double zeta = 0.;
    double d = 0.;

    // Initialize variables for estimation of ||A|| and cond(A)

    double normA2 = alpha * alpha;
    double maxrbar = 0.;
    double minrbar = 1.e+100;
    m_normA = std::sqrt(normA2);
    m_condA = 1.;
    m_normx = 0.;

    // Items for use in stopping rules.
    double normb = beta;
    m_istop = 0;
    double ctol = 0.;
    if (m_conlim > 0.)
        ctol = 1. / m_conlim;
    m_normr = beta;

    // Reverse the order here from the original matlab code because
    // there was an error on return when arnorm==0
    m_normar = alpha * beta;
    if (m_normar == 0.) {
        m_stopMessage = m_msg[0];
        if (m_output)
            *m_output << m_stopMessage;
        return EigenUtils::convertToVectorD(x); //, istop, itn, normr, normar, normA, condA, normx
    }

    double test1, test2, test3, t1, rtol;
    if (m_output) {
        *m_output << "\n";
        *m_output << m_hdg1 << " " << m_hdg2 << "\n";
        test1 = 1.;
        test2 = alpha / beta;
        *m_output << m_itn << " " << x(0) << " " << m_normr << " " << m_normar
                << " " << test1 << " " << test2 << "\n";
    }

    // Initialize variables used in the loop
    double chat, shat, alphahat, rhoold, c, s, thetanew, rhobarold, zetaold,
            thetabar, rhotemp, betaacute, betacheck, betahat, thetatildeold,
            ctildeold, stildeold, rhotildeold, taud, betadtaud;

    // Test time of an iteration
    std::clock_t start;
    double duration;

    // Main iteration loop.
    while (m_itn < m_maxiter) {
        m_itn++;

        // Test time of an iteration
        start = std::clock();

        // Perform the next step of the bidiagonalization to obtain the
        // next  beta, u, alpha, v.  These satisfy the relations
        //         beta*u  =  a*v   -  alpha*u,
        //        alpha*v  =  A'*u  -  beta*v.

        u = (EigenA * v) - (u * alpha);
        beta = u.norm();

        if (beta > 0) {
            u = u / beta;
//            v = (At * u) - (v * beta);
            v = (EigenAt * u) - (v * beta);
            alpha = v.norm();
            if (alpha > 0)
                v = v / alpha;
        }

        // At this point, beta = beta_{k+1}, alpha = alpha_{k+1}.

        // Construct rotation Qhat_{k,2k+1}.

        _sym_ortho(alphabar, m_damp, chat, shat, alphahat);

        // Use a plane rotation (Q_i) to turn B_i to R_i

        rhoold = rho;
        _sym_ortho(alphahat, beta, c, s, rho);
        thetanew = s * alpha;
        alphabar = c * alpha;

        // Use a plane rotation (Qbar_i) to turn R_i^T to R_i^bar

        rhobarold = rhobar;
        zetaold = zeta;
        thetabar = sbar * rho;
        rhotemp = cbar * rho;
        _sym_ortho(cbar * rho, thetanew, cbar, sbar, rhobar);
        zeta = cbar * zetabar;
        zetabar = -sbar * zetabar;

        // Update h, h_hat, x.

        hbar = h - (hbar * (thetabar * rho / (rhoold * rhobarold)));
        x = x + (hbar * (zeta / (rho * rhobar)));
        h = v - (h * (thetanew / rho));

        // Estimate of ||r||.

        // Apply rotation Qhat_{k,2k+1}.
        betaacute = chat * betadd;
        betacheck = -shat * betadd;

        // Apply rotation Q_{k,k+1}.
        betahat = c * betaacute;
        betadd = -s * betaacute;

        // Apply rotation Qtilde_{k-1}.
        // betad = betad_{k-1} here.

        thetatildeold = thetatilde;
        _sym_ortho(rhodold, thetabar, ctildeold, stildeold, rhotildeold);
        thetatilde = stildeold * rhobar;
        rhodold = ctildeold * rhobar;
        betad = -stildeold * betad + ctildeold * betahat;

        // betad   = betad_k here.
        // rhodold = rhod_k  here.

        tautildeold = (zetaold - thetatildeold * tautildeold) / rhotildeold;
        taud = (zeta - thetatilde * tautildeold) / rhodold;
        d = d + betacheck * betacheck;
        betadtaud = betad - taud;
        m_normr = std::sqrt(d + betadtaud * betadtaud + betadd * betadd);

        // Estimate ||A||.
        normA2 = normA2 + beta * beta;
        m_normA = std::sqrt(normA2);
        normA2 = normA2 + alpha * alpha;

        // Estimate cond(A).
        maxrbar = std::max(maxrbar, rhobarold);
        if (m_itn > 1)
            minrbar = std::min(minrbar, rhobarold);
        m_condA = std::max(maxrbar, rhotemp) / std::min(minrbar, rhotemp);

        // Test for convergence.

        // Compute norms for convergence testing.
        m_normar = fabs(zetabar);
        m_normx = x.norm();

        // Now use these norms to estimate certain other quantities,
        // some of which will be small near a solution.

        test1 = m_normr / normb;
        if ((m_normA * m_normr) != 0.)
            test2 = m_normar / (m_normA * m_normr);
        else
            test2 = std::numeric_limits<double>::infinity();
        test3 = 1 / m_condA;
        t1 = test1 / (1 + m_normA * m_normx / normb);
        rtol = m_btol + m_atol * m_normA * m_normx / normb;

        // The following tests guard against extremely small values of
        // atol, btol or ctol.  (The user may have set any or all of
        // the parameters atol, btol, conlim  to 0.)
        // The effect is equivalent to the normAl tests using
        // atol = eps,  btol = eps,  conlim = 1/eps.

        if (m_itn >= m_maxiter)
            m_istop = 7;
        if (1 + test3 <= 1)
            m_istop = 6;
        if (1 + test2 <= 1)
            m_istop = 5;
        if (1 + t1 <= 1)
            m_istop = 4;

        // Allow for tolerances set by the user.

        if (test3 <= ctol)
            m_istop = 3;
        if (test2 <= m_atol)
            m_istop = 2;
        if (test1 <= rtol)
            m_istop = 1;

        // Test time of an iteration
        duration = (std::clock() - start) / (double) CLOCKS_PER_SEC;

        // See if it is time to print something.

        if (m_output) {
            if ((n <= 40) or (m_itn <= 10) or (m_itn >= m_maxiter - 10)
                    or (m_itn % 10 == 0) or (test3 <= 1.1 * ctol)
                    or (test2 <= 1.1 * m_atol) or (test1 <= 1.1 * rtol)
                    or (m_istop != 0)) {
                if (pcount >= m_pfreq) {
                    pcount = 0;
                    *m_output << "\n";
                    *m_output << m_hdg1 << m_hdg2 << "\n";
                }
                pcount++;
                *m_output << m_itn;
                *m_output << " ";
                *m_output << x(0);
                *m_output << " ";
                *m_output << m_normr;
                *m_output << " ";
                *m_output << m_normar;
                *m_output << " ";
                *m_output << test1;
                *m_output << " ";
                *m_output << test2;
                *m_output << " ";
                *m_output << m_normA;
                *m_output << " ";
                *m_output << m_condA;
                *m_output << " ";
                *m_output << duration;
                *m_output << "s";
                *m_output << "\n";
            }
        }

        if (m_istop > 0)
            break;

    }

    // Stopping reason
    m_stopMessage = m_msg[m_istop];

    // Print the stopping condition.

    if (m_output) {
        *m_output << "\n";
        *m_output << "LSMR finished" << "\n";
        *m_output << m_stopMessage << "\n";
        *m_output << "istop = " << m_istop << "  normr = " << m_normr << "\n";
        *m_output << "  normA = " << m_normA << " normAr = " << m_normar
                << "\n";
        *m_output << "itn = " << m_itn << " condA = " << m_condA << "\n";
        *m_output << "  normx =" << m_normx << "\n";
    }

    return EigenUtils::convertToVectorD(x); //, istop, itn, normr, normar, normA, condA, normx
}

double LSMRSolver::getToleranceA() const {
    return m_atol;
}

void LSMRSolver::setToleranceA(double atol) {
    m_atol = atol;
}

double LSMRSolver::getToleranceB() const {
    return m_btol;
}

void LSMRSolver::setToleranceB(double btol) {
    m_btol = btol;
}

double LSMRSolver::getConditionNumberLimit() const {
    return m_conlim;
}

void LSMRSolver::setConditionNumberLimit(double conlim) {
    m_conlim = conlim;
}

double LSMRSolver::getDamping() const {
    return m_damp;
}

void LSMRSolver::setDamping(double damp) {
    m_damp = damp;
}

size_t LSMRSolver::getMaximumIerations() const {
    return m_maxiter;
}

void LSMRSolver::setMaximumIerations(size_t maxiter) {
    m_maxiter = maxiter;
}

ElemUtils::Formatter& LSMRSolver::getOutput() const {
    return *m_output;
}

void LSMRSolver::setOutput(ElemUtils::Formatter& output) {
    m_output = &output;
}

double LSMRSolver::getConditionNumber() const {
    return m_condA;
}

unsigned int LSMRSolver::getStoppingCase() const {
    return m_istop;
}

size_t LSMRSolver::getNumberIterations() const {
    return m_itn;
}

double LSMRSolver::getNormA() const {
    return m_normA;
}

double LSMRSolver::getNormAr() const {
    return m_normar;
}

double LSMRSolver::getNormr() const {
    return m_normr;
}

double LSMRSolver::getNormX() const {
    return m_normx;
}

const std::string& LSMRSolver::getStoppingMessage() const {
    return m_stopMessage;
}

void LSMRSolver::_sym_ortho(double a, double b, double& c, double& s,
        double& r) const {
    if (b == 0) {
        c = (a > 0) - (a < 0);
        s = 0;
        r = fabs(a);
    } else if (a == 0) {
        c = 0;
        s = (b > 0) - (b < 0);
        r = fabs(b);
    } else if (fabs(b) > fabs(a)) {
        double tau = a / b;
        s = ((b > 0) - (b < 0)) / std::sqrt(1. + tau * tau);
        c = s * tau;
        r = b / s;
    } else {
        double tau = b / a;
        c = ((a > 0) - (a < 0)) / std::sqrt(1. + tau * tau);
        s = c * tau;
        r = a / c;
    }
}

} /* namespace NumA */
