/*
 * MatrixComplexD.cpp
 *
 *  Created on: Jan 16, 2016
 *      Author: Pawel Sznajder (IPNO)
 */

#include "../../../../include/NumA/linear_algebra/matrix_complex/MatrixComplexD.h"

#include <stdexcept>
#include <vector>

namespace NumA {

MatrixComplexD::MatrixComplexD() {

    throw std::runtime_error(
            "MatrixComplexD::MatrixComplexD(): Can not set matrix of dimensions equal zero");

    m_n = 0;
    m_m = 0;
    m_elements.clear();
}

MatrixComplexD::MatrixComplexD(const unsigned int n, const unsigned int m) {

    if (n <= 0 || m <= 0)
        throw std::runtime_error(
                "MatrixComplexD::MatrixComplexD(): Can not set matrix of dimensions equal zero or negative");

    this->m_n = n;
    this->m_m = m;

    m_elements = std::vector< std::vector<std::complex<double> > >(n, std::vector<std::complex<double> >(m, std::complex<double>(0., 0.)));
}

MatrixComplexD::MatrixComplexD(const MatrixComplexD& v) {

    this->m_n = v.m_n;
    this->m_m = v.m_m;

    this->m_elements = v.m_elements;
}

MatrixComplexD::~MatrixComplexD() {
}

MatrixComplexD MatrixComplexD::Conjunct() const {

    MatrixComplexD r(this->m_n, this->m_m);

    for (unsigned int i = 0; i < this->m_n; i++) {
        for (unsigned int j = 0; j < this->m_m; j++) {
            r.SetElement(i, j, std::conj(this->m_elements.at(i).at(j)));
        }
    }

    return r;
}

MatrixComplexD MatrixComplexD::Mult(const MatrixComplexD& rhs) const {

    if (this->m_n != rhs.m_m) {
        throw std::runtime_error(
                "MatrixComplexD::Mult(): Width of first matrix is not equal height of second matrix");
    }

    MatrixComplexD r(this->m_n, this->m_m);

    for (unsigned int i = 0; i < this->m_n; i++) {
        for (unsigned int j = 0; j < this->m_m; j++) {
            r.SetElement(i, j, std::complex<double>(0., 0.));

            for (unsigned int k = 0; k < this->m_n; k++) {
                r.SetElement(i, j,
                        r.GetElement(i, j)
                                + this->m_elements.at(i).at(k) * rhs.m_elements.at(k).at(j));
            }
        }
    }

    return r;
}

void MatrixComplexD::Clear() {

    for (unsigned int i = 0; i < this->m_n; i++) {
        for (unsigned int j = 0; j < this->m_m; j++) {
            this->m_elements.at(i).at(j) = std::complex<double>(0., 0.);
        }
    }
}

void MatrixComplexD::MakeUnit() {

    if (this->m_n != this->m_m)
        throw std::runtime_error(
                "MatrixComplexD::MakeRealUnit(): Unit matrix only for square ones");

    for (unsigned int i = 0; i < this->m_n; i++) {
        this->m_elements.at(i).at(i) = std::complex<double>(1., 0.);
    }
}

void MatrixComplexD::operator =(const MatrixComplexD& rhs) {

    if (this == &rhs)
        return;

    this->CompareNElements(rhs);

    for (unsigned int i = 0; i < this->m_n; i++) {
        for (unsigned int j = 0; j < this->m_m; j++) {
            this->m_elements.at(i).at(j) = rhs.m_elements.at(i).at(j);
        }
    }
}

MatrixComplexD MatrixComplexD::operator +(const MatrixComplexD& rhs) {

    this->CompareNElements(rhs);

    MatrixComplexD r(this->m_n, this->m_m);

    for (unsigned int i = 0; i < this->m_n; i++) {
        for (unsigned int j = 0; j < this->m_m; j++) {
            r.SetElement(i, j, this->m_elements.at(i).at(j) + rhs.m_elements.at(i).at(j));
        }
    }

    return r;
}

MatrixComplexD MatrixComplexD::operator -(const MatrixComplexD& rhs) {

    this->CompareNElements(rhs);

    MatrixComplexD r(this->m_n, this->m_m);

    for (unsigned int i = 0; i < this->m_n; i++) {
        for (unsigned int j = 0; j < this->m_m; j++) {
            r.SetElement(i, j, this->m_elements.at(i).at(j) - rhs.m_elements.at(i).at(j));
        }
    }

    return r;
}

MatrixComplexD MatrixComplexD::operator *(const double& rhs) {

    MatrixComplexD r(this->m_n, this->m_m);

    for (unsigned int i = 0; i < this->m_n; i++) {
        for (unsigned int j = 0; j < this->m_m; j++) {
            r.SetElement(i, j, rhs * (this->m_elements.at(i).at(j)));
        }
    }

    return r;
}

MatrixComplexD operator *(const double& lhs, MatrixComplexD& rhs) {
    return rhs * lhs;
}

MatrixComplexD MatrixComplexD::operator *(const std::complex<double>& rhs) {

    MatrixComplexD r(this->m_n, this->m_m);

    for (unsigned int i = 0; i < this->m_n; i++) {
        for (unsigned int j = 0; j < this->m_m; j++) {
            r.SetElement(i, j, rhs * (this->m_elements.at(i).at(j)));
        }
    }

    return r;
}

MatrixComplexD operator *(const std::complex<double>& lhs,
        MatrixComplexD& rhs) {
    return rhs * lhs;
}

unsigned int MatrixComplexD::GetNElements(const Dim v) const{

    switch (v) {

    case X: {

        return m_n;
        break;
    }

    case Y: {

        return m_m;
        break;
    }

    default: {

        throw std::runtime_error(
                "MatrixComplexD::GetDimension(): Illegal index");
        break;
    }

    }

    return 0;
}

void MatrixComplexD::SetElement(const unsigned int i, const unsigned int j, const std::complex<double> v) {

    if (i >= this->m_n || j >= this->m_m)
        throw std::runtime_error("MatrixComplexD::SetElement(): Illegal index");

    this->m_elements.at(i).at(j) = v;
}

std::complex<double> MatrixComplexD::GetElement(const unsigned int i, const unsigned int j) const {

    if (i >= this->m_n || j >= this->m_m)
        throw std::runtime_error("MatrixComplexD::GetElement(): Illegal index");

    return m_elements.at(i).at(j);
}

void MatrixComplexD::CompareNElements(const MatrixComplexD& v) const {

    if (this->m_n != v.GetNElements(X) || this->m_m != v.GetNElements(Y))
        throw std::runtime_error(
                "MatrixComplexD::CompareDimension(): Two complex matrices have different dimensions");

}

} /* namespace NumA */
