#include "../../../../include/NumA/linear_algebra/vector/VectorD.h"

#include <cmath>
#include <ElementaryUtils/logger/CustomException.h>
#include <ElementaryUtils/string_utils/Formatter.h>

#include "../../../../include/NumA/linear_algebra/matrix/MatrixD.h"
#include "../../../../include/NumA/linear_algebra/vector/Vector4D.h"

namespace NumA {

VectorD::VectorD() {
}

VectorD::VectorD(size_t size) {
    m_vector.assign(size, 0.);
}

VectorD::VectorD(size_t size, double initValue) {
    m_vector.assign(size, initValue);
}

VectorD::VectorD(const std::vector<double> &vector) :
        m_vector(vector) {

}

VectorD::VectorD(const VectorD& vector) {
    m_vector = vector.toStdVector();
}

VectorD::VectorD(const Vector2D& vector) {
    m_vector.assign(2, vector.getX());
    m_vector[1] = vector.getY();
}

VectorD::VectorD(const Vector3D& vector) {
    m_vector.assign(3, vector.getX());
    m_vector[1] = vector.getY();
    m_vector[2] = vector.getZ();
}

VectorD::VectorD(const Vector4D& vector) {
    m_vector.assign(4, vector.getX());
    m_vector[1] = vector.getY();
    m_vector[2] = vector.getZ();
    m_vector[3] = vector.getE();
}

VectorD::~VectorD() {
}

void VectorD::push_back(double value) {
    m_vector.push_back(value);
}

void VectorD::resize(size_t n, double value) {
    m_vector.resize(n, value);
}

void VectorD::assign(size_t n, double value) {
    m_vector.assign(n, value);
}

size_t VectorD::size() const {
    return m_vector.size();
}

double VectorD::operator *(const VectorD &rhs) const {
    double result = 0.;

    unsigned int vectorSize = m_vector.size();

    // cannot perform multiplication if the two vector's size are different
    if (vectorSize != rhs.size()) {
        throw ElemUtils::CustomException("VectorD", __func__,
                "Vectors have different size.");
    }

    for (unsigned int i = 0; i < vectorSize; i++) {
        result += (m_vector[i] * rhs[i]);
    }

    return result;
}

VectorD NumA::VectorD::operator -(const VectorD& rhs) const {
    // cannot perform subtraction if the two vector's size are different
    if (size() != rhs.size()) {
        throw ElemUtils::CustomException("VectorD", __func__,
                "Vectors have different size.");
    }

    VectorD result(size());
    for (unsigned int i = 0; i < size(); i++) {
        result[i] = (m_vector[i] - rhs[i]);
    }
    return result;
}

VectorD VectorD::operator +(const VectorD& rhs) const {
    // cannot perform addition if the two vector's size are different
    if (size() != rhs.size()) {
        throw ElemUtils::CustomException("VectorD", __func__,
                "Vectors have different size.");
    }

    VectorD result(size());
    for (unsigned int i = 0; i < size(); i++) {
        result[i] = (m_vector[i] + rhs[i]);
    }
    return result;
}

VectorD VectorD::operator *(double rhs) const {
    VectorD result(size());
    for (unsigned int i = 0; i < size(); i++) {
        result[i] = (m_vector[i] * rhs);
    }
    return result;
}

VectorD VectorD::operator +(double rhs) const {
    VectorD result(size());
    for (unsigned int i = 0; i < size(); i++) {
        result[i] = (m_vector[i] + rhs);
    }
    return result;
}

VectorD VectorD::operator -(double rhs) const {
    VectorD result(size());
    for (unsigned int i = 0; i < size(); i++) {
        result[i] = (m_vector[i] - rhs);
    }
    return result;
}

VectorD VectorD::operator /(double rhs) const {
    VectorD result(size());
    for (unsigned int i = 0; i < size(); i++) {
        result[i] = (m_vector[i] / rhs);
    }
    return result;
}

double NumA::VectorD::norm() const {
    double result = (*this) * (*this);
    result = std::sqrt(result);
    return result;
}

double& VectorD::operator [](size_t n) {
    return m_vector[n];
}

double& VectorD::at(size_t n) {
    return m_vector.at(n);
}

const double& VectorD::at(size_t n) const {
    return m_vector.at(n);
}

const double& VectorD::operator [](size_t n) const {
    return m_vector[n];
}

VectorD VectorD::sub(size_t startPos, size_t endPos) const {
    VectorD result(static_cast<int>(endPos) - static_cast<int>(startPos));

    for (size_t i = 0; i < result.size(); i++) {
        result[i] = m_vector[i + startPos];
    }

    return result;
}

const std::vector<double>& VectorD::toStdVector() const {
    return m_vector;
}

VectorD VectorD::concatenate(const VectorD& V1, const VectorD& V2) {
    std::vector<double> v = V1.toStdVector();
    std::vector<double> v2 = V2.toStdVector();
    v.insert(v.end(), v2.begin(), v2.end());
    VectorD V(v);
    return V;
}

MatrixD VectorD::toMatrix() {
    MatrixD matrix(*this);
    return matrix;
}

std::string VectorD::toString() const {
    ElemUtils::Formatter formatter;

    for (unsigned int i = 0; i < m_vector.size(); i++) {
        formatter << at(i);
        formatter << " ";
    }

    formatter << "\n";

    return formatter.str();
}

} /* namespace NumA */
