/*
 * NeuralNetworkTypeRegistry.cpp
 *
 *  Created on: Apr 23, 2016
 *      Author: Pawel Sznajder
 */

#include "../../../include/NumA/neural_network/NeuralNetworkTypeRegistry.h"

#include <ElementaryUtils/logger/CustomException.h>
#include <ElementaryUtils/string_utils/Formatter.h>
#include <string>
#include <utility>

namespace NumA {

NeuralNetworkTypeRegistry::NeuralNetworkTypeRegistry() {

    m_activationFunctions.clear();
    m_combinationFunctions.clear();
    m_scalingFunctions.clear();
    m_trainingFunctions.clear();
}

NeuralNetworkTypeRegistry::~NeuralNetworkTypeRegistry() {

    std::map<ActivationFunctionType::Type, ActivationFunction*>::iterator itActivationFunctions;

    for (itActivationFunctions = m_activationFunctions.begin();
            itActivationFunctions != m_activationFunctions.end();
            itActivationFunctions++) {
        if (itActivationFunctions->second) {
            delete (itActivationFunctions->second);
            (itActivationFunctions->second) = 0;
        }
    }

    std::map<CombinationFunctionType::Type, CombinationFunction*>::iterator itCombinationFunctions;

    for (itCombinationFunctions = m_combinationFunctions.begin();
            itCombinationFunctions != m_combinationFunctions.end();
            itCombinationFunctions++) {
        if (itCombinationFunctions->second) {
            delete (itCombinationFunctions->second);
            (itCombinationFunctions->second) = 0;
        }
    }

    std::map<ScalingFunctionType::Type, ScalingFunction*>::iterator itScalingFunctions;

    for (itScalingFunctions = m_scalingFunctions.begin();
            itScalingFunctions != m_scalingFunctions.end();
            itScalingFunctions++) {
        if (itScalingFunctions->second) {
            delete (itScalingFunctions->second);
            (itScalingFunctions->second) = 0;
        }
    }

    std::map<TrainingFunctionType::Type, TrainingFunction*>::iterator itTrainingFunctions;

    for (itTrainingFunctions = m_trainingFunctions.begin();
            itTrainingFunctions != m_trainingFunctions.end();
            itTrainingFunctions++) {
        if (itTrainingFunctions->second) {
            delete (itTrainingFunctions->second);
            (itTrainingFunctions->second) = 0;
        }
    }

    if (p_instance) {
        delete p_instance;
        p_instance = 0;
    }
}

NeuralNetworkTypeRegistry* NeuralNetworkTypeRegistry::p_instance = 0;

NeuralNetworkTypeRegistry* NeuralNetworkTypeRegistry::getInstance() {

    if (p_instance == 0) {
        p_instance = new NeuralNetworkTypeRegistry();
    }

    return p_instance;
}

unsigned int NeuralNetworkTypeRegistry::registerActivationFunction(
        const ActivationFunctionType::Type type,
        ActivationFunction* const object) {

    if (type == ActivationFunctionType::UNDEFINED) {
        throw ElemUtils::CustomException("NeuralNetworkTypeRegistry", __func__,
                "Undefined type");
    }

    if (m_activationFunctions.find(type) != m_activationFunctions.end()) {
        throw ElemUtils::CustomException("NeuralNetworkTypeRegistry", __func__,
                "Object already registered");
    }

    m_activationFunctions.insert(std::make_pair(type, object));

    return m_activationFunctions.size();
}

unsigned int NeuralNetworkTypeRegistry::registerCombinationFunction(
        const CombinationFunctionType::Type type,
        CombinationFunction* const object) {

    if (type == CombinationFunctionType::UNDEFINED) {
        throw ElemUtils::CustomException("NeuralNetworkTypeRegistry", __func__,
                "Undefined type");
    }

    if (m_combinationFunctions.find(type) != m_combinationFunctions.end()) {
        throw ElemUtils::CustomException("NeuralNetworkTypeRegistry", __func__,
                "Object already registered");
    }

    m_combinationFunctions.insert(std::make_pair(type, object));

    return m_combinationFunctions.size();
}

unsigned int NeuralNetworkTypeRegistry::registerScalingFunction(
        const ScalingFunctionType::Type type, ScalingFunction* const object) {

    if (type == ScalingFunctionType::UNDEFINED) {
        throw ElemUtils::CustomException("NeuralNetworkTypeRegistry", __func__,
                "Undefined type");
    }

    if (m_scalingFunctions.find(type) != m_scalingFunctions.end()) {
        throw ElemUtils::CustomException("NeuralNetworkTypeRegistry", __func__,
                "Object already registered");
    }

    m_scalingFunctions.insert(std::make_pair(type, object));

    return m_scalingFunctions.size();
}

unsigned int NeuralNetworkTypeRegistry::registerTrainingFunction(
        const TrainingFunctionType::Type type, TrainingFunction* const object) {

    if (type == TrainingFunctionType::UNDEFINED) {
        throw ElemUtils::CustomException("NeuralNetworkTypeRegistry", __func__,
                "Undefined type");
    }

    if (m_trainingFunctions.find(type) != m_trainingFunctions.end()) {
        throw ElemUtils::CustomException("NeuralNetworkTypeRegistry", __func__,
                "Object already registered");
    }

    m_trainingFunctions.insert(std::make_pair(type, object));

    return m_trainingFunctions.size();
}

ActivationFunction* NeuralNetworkTypeRegistry::getActivationFunction(
        const ActivationFunctionType::Type type) const {

    if (type == ActivationFunctionType::UNDEFINED) {
        throw ElemUtils::CustomException("NeuralNetworkTypeRegistry", __func__,
                "Undefined type");
    }

    std::map<ActivationFunctionType::Type, ActivationFunction*>::const_iterator it =
            m_activationFunctions.find(type);

    if (it == m_activationFunctions.end()) {
        throw ElemUtils::CustomException("NeuralNetworkTypeRegistry", __func__,
                ElemUtils::Formatter() << "Object not registered, type "
                        << ActivationFunctionType(type).toString());
    }

    return it->second;
}

CombinationFunction* NeuralNetworkTypeRegistry::getCombinationFunction(
        const CombinationFunctionType::Type type) const {

    if (type == CombinationFunctionType::UNDEFINED) {
        throw ElemUtils::CustomException("NeuralNetworkTypeRegistry", __func__,
                "Undefined type");
    }

    std::map<CombinationFunctionType::Type, CombinationFunction*>::const_iterator it =
            m_combinationFunctions.find(type);

    if (it == m_combinationFunctions.end()) {
        throw ElemUtils::CustomException("NeuralNetworkTypeRegistry", __func__,
                ElemUtils::Formatter() << "Object not registered, type "
                        << CombinationFunctionType(type).toString());
    }

    return it->second;
}

ScalingFunction* NeuralNetworkTypeRegistry::getScalingFunction(
        const ScalingFunctionType::Type type) const {

    if (type == ScalingFunctionType::UNDEFINED) {
        throw ElemUtils::CustomException("NeuralNetworkTypeRegistry", __func__,
                "Undefined type");
    }

    std::map<ScalingFunctionType::Type, ScalingFunction*>::const_iterator it =
            m_scalingFunctions.find(type);

    if (it == m_scalingFunctions.end()) {
        throw ElemUtils::CustomException("NeuralNetworkTypeRegistry", __func__,
                ElemUtils::Formatter() << "Object not registered, type "
                        << ScalingFunctionType(type).toString());
    }

    return it->second;
}

TrainingFunction* NeuralNetworkTypeRegistry::getTrainingFunction(
        const TrainingFunctionType::Type type) const {

    if (type == TrainingFunctionType::UNDEFINED) {
        throw ElemUtils::CustomException("NeuralNetworkTypeRegistry", __func__,
                "Undefined type");
    }

    std::map<TrainingFunctionType::Type, TrainingFunction*>::const_iterator it =
            m_trainingFunctions.find(type);

    if (it == m_trainingFunctions.end()) {
        throw ElemUtils::CustomException("NeuralNetworkTypeRegistry", __func__,
                ElemUtils::Formatter() << "Object not registered, type "
                        << TrainingFunctionType(type).toString());
    }

    return it->second;
}

}
